r"""Defines some utilities function to analyze a ``torch.nn.Module``.

Inspired from https://github.com/PyTorchLightning/pytorch-
lightning/blob/master/pytorch_lightning/core/memory.py
"""

__all__ = ["ModelSummary"]

from collections import OrderedDict

import torch
from tabulate import tabulate
from torch.nn import Module

from gravitorch.nn.utils.summary import (
    ModuleSummary,
    multiline_format_dtype,
    multiline_format_size,
)
from gravitorch.utils.format import convert_human_readable_count


class ModelSummary:
    r"""Generates a summary of all layers in a ``torch.nn.Module``.

    Args:
        model (``torch.nn.Module``): The model to summarize.
        mode: Can be one of

             - `top` (default): only the top-level modules will be
                recorded (the children of the root module)
             - `full`: summarizes all layers and their submodules
                in the root module

    The string representation of this summary prints a table with
    columns containing the name, type and number of parameters for
    each layer.

    The root module may also have a function ``get_dummy_input`` as
    shown in the example below. If present, the root module will be
    called with it as input to determine the intermediate input and
    output shapes and data types of all layers. Supported are tensors
    and nested lists and tuples of tensors. All other types of inputs
    will be skipped and show as `?` in the summary table. The summary
    will also display `?` for layers not used in the forward pass.

    Example::

        >>> import torch
        >>> from torch import nn
        >>> from gravitorch.models.utils import ModelSummary
        >>> model = torch.nn.Linear(4, 5)
        >>> print(ModelSummary(model, mode='top'))  # doctest: +NORMALIZE_WHITESPACE
        ╒════╤═══════════════╤════════╤══════════╤════════════════╕
        │    │ Name          │ Type   │   Params │   Learn Params │
        ╞════╪═══════════════╪════════╪══════════╪════════════════╡
        │  0 │ [root module] │ Linear │     25   │           25   │
        ╘════╧═══════════════╧════════╧══════════╧════════════════╛
         - 25         Learnable params
         - 0          Non-learnable params
         - 25         Total params

        >>> from gravitorch.models.networks import BetaMLP
        >>> from gravitorch.nn import ConcatFusion
        >>> class MyNetwork(nn.Module):
        ...     def __init__(self):
        ...         super().__init__()
        ...         self.fusion = ConcatFusion()
        ...         self.decoder = BetaMLP(input_size=20, hidden_sizes=(16, 7))
        ...
        ...     def forward(self, x1: torch.Tensor, x2: torch.Tensor, x3: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        ...         return self.decoder(self.fusion(x1, x2, x3)), x3
        ...
        ...     def get_dummy_input(self, batch_size: int = 1) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        ...         return (
        ...             torch.randn(batch_size, 10),
        ...             torch.randn(batch_size, 2),
        ...             torch.randn(batch_size, 8),
        ...         )
        ...
        >>> model = MyNetwork()
        >>> ModelSummary(model, mode='top')  # doctest: +NORMALIZE_WHITESPACE
        ╒════╤═══════════════╤══════════════╤══════════╤════════════════╤════════════╤═══════════════╤═════════════╤═══════════════╕  # noqa: E501,B950
        │    │ Name          │ Type         │   Params │   Learn Params │ In sizes   │ In dtype      │ Out sizes   │ Out dtype     │  # noqa: E501,B950
        ╞════╪═══════════════╪══════════════╪══════════╪════════════════╪════════════╪═══════════════╪═════════════╪═══════════════╡  # noqa: E501,B950
        │  0 │ [root module] │ MyNetwork    │    455   │          455   │ (1, 10)    │ torch.float32 │ (1, 7)      │ torch.float32 │  # noqa: E501,B950
        │    │               │              │          │                │ (1, 2)     │ torch.float32 │ (1, 8)      │ torch.float32 │  # noqa: E501,B950
        │    │               │              │          │                │ (1, 8)     │ torch.float32 │             │               │  # noqa: E501,B950
        ├────┼───────────────┼──────────────┼──────────┼────────────────┼────────────┼───────────────┼─────────────┼───────────────┤  # noqa: E501,B950
        │  1 │ fusion        │ ConcatFusion │      0   │            0   │ (1, 10)    │ torch.float32 │ (1, 20)     │ torch.float32 │  # noqa: E501,B950
        │    │               │              │          │                │ (1, 2)     │ torch.float32 │             │               │  # noqa: E501,B950
        │    │               │              │          │                │ (1, 8)     │ torch.float32 │             │               │  # noqa: E501,B950
        ├────┼───────────────┼──────────────┼──────────┼────────────────┼────────────┼───────────────┼─────────────┼───────────────┤  # noqa: E501,B950
        │  2 │ decoder       │ BetaMLP      │    455   │          455   │ (1, 20)    │ torch.float32 │ (1, 7)      │ torch.float32 │  # noqa: E501,B950
        ╘════╧═══════════════╧══════════════╧══════════╧════════════════╧════════════╧═══════════════╧═════════════╧═══════════════╛  # noqa: E501,B950
         - 455        Learnable params
         - 0          Non-learnable params
         - 455        Total params

        >>> ModelSummary(model, mode='full')  # doctest: +NORMALIZE_WHITESPACE
        ╒════╤════════════════════════╤══════════════╤══════════╤════════════════╤════════════╤═══════════════╤═════════════╤═══════════════╕  # noqa: E501,B950
        │    │ Name                   │ Type         │   Params │   Learn Params │ In sizes   │ In dtype      │ Out sizes   │ Out dtype     │  # noqa: E501,B950
        ╞════╪════════════════════════╪══════════════╪══════════╪════════════════╪════════════╪═══════════════╪═════════════╪═══════════════╡  # noqa: E501,B950
        │  0 │ [root module]          │ MyNetwork    │    455   │          455   │ (1, 10)    │ torch.float32 │ (1, 7)      │ torch.float32 │  # noqa: E501,B950
        │    │                        │              │          │                │ (1, 2)     │ torch.float32 │ (1, 8)      │ torch.float32 │  # noqa: E501,B950
        │    │                        │              │          │                │ (1, 8)     │ torch.float32 │             │               │  # noqa: E501,B950
        ├────┼────────────────────────┼──────────────┼──────────┼────────────────┼────────────┼───────────────┼─────────────┼───────────────┤  # noqa: E501,B950
        │  1 │ fusion                 │ ConcatFusion │      0   │            0   │ (1, 10)    │ torch.float32 │ (1, 20)     │ torch.float32 │  # noqa: E501,B950
        │    │                        │              │          │                │ (1, 2)     │ torch.float32 │             │               │  # noqa: E501,B950
        │    │                        │              │          │                │ (1, 8)     │ torch.float32 │             │               │  # noqa: E501,B950
        ├────┼────────────────────────┼──────────────┼──────────┼────────────────┼────────────┼───────────────┼─────────────┼───────────────┤  # noqa: E501,B950
        │  2 │ decoder                │ BetaMLP      │    455   │          455   │ (1, 20)    │ torch.float32 │ (1, 7)      │ torch.float32 │  # noqa: E501,B950
        ├────┼────────────────────────┼──────────────┼──────────┼────────────────┼────────────┼───────────────┼─────────────┼───────────────┤  # noqa: E501,B950
        │  3 │ decoder.layers         │ Sequential   │    455   │          455   │ (1, 20)    │ torch.float32 │ (1, 7)      │ torch.float32 │  # noqa: E501,B950
        ├────┼────────────────────────┼──────────────┼──────────┼────────────────┼────────────┼───────────────┼─────────────┼───────────────┤  # noqa: E501,B950
        │  4 │ decoder.layers.linear1 │ Linear       │    336   │          336   │ (1, 20)    │ torch.float32 │ (1, 16)     │ torch.float32 │  # noqa: E501,B950
        ├────┼────────────────────────┼──────────────┼──────────┼────────────────┼────────────┼───────────────┼─────────────┼───────────────┤  # noqa: E501,B950
        │  5 │ decoder.layers.relu1   │ ReLU         │      0   │            0   │ (1, 16)    │ torch.float32 │ (1, 16)     │ torch.float32 │  # noqa: E501,B950
        ├────┼────────────────────────┼──────────────┼──────────┼────────────────┼────────────┼───────────────┼─────────────┼───────────────┤  # noqa: E501,B950
        │  6 │ decoder.layers.linear2 │ Linear       │    119   │          119   │ (1, 16)    │ torch.float32 │ (1, 7)      │ torch.float32 │  # noqa: E501,B950
        ╘════╧════════════════════════╧══════════════╧══════════╧════════════════╧════════════╧═══════════════╧═════════════╧═══════════════╛  # noqa: E501,B950
         - 455        Learnable params
         - 0          Non-learnable params
         - 455        Total params
    """

    MODE_TOP = "top"
    MODE_FULL = "full"
    MODE_DEFAULT = MODE_TOP
    MODES = [MODE_FULL, MODE_TOP]

    def __init__(self, model: Module, mode: str = MODE_DEFAULT):
        self._model = model
        self._mode = mode
        self._layer_summary = self.summarize()

    @property
    def named_modules(self) -> list[tuple[str, Module]]:
        modules = [("[root module]", self._model)]
        if self._mode == ModelSummary.MODE_FULL:
            # All the modules including the root module.
            modules.extend(list(self._model.named_modules())[1:])
            return modules
        elif self._mode == ModelSummary.MODE_TOP:
            # The children are the top-level modules.
            modules.extend(list(self._model.named_children()))
            return modules
        raise ValueError(f"Incorrect mode: {self._mode}. The valid modes are {self.MODES}.")

    @property
    def layer_names(self) -> tuple[str, ...]:
        return tuple(self._layer_summary.keys())

    @property
    def layer_types(self) -> tuple[str, ...]:
        return tuple(layer.layer_type for layer in self._layer_summary.values())

    @property
    def in_sizes(self) -> tuple:
        return tuple(layer.in_size for layer in self._layer_summary.values())

    @property
    def out_sizes(self) -> tuple:
        return tuple(layer.out_size for layer in self._layer_summary.values())

    @property
    def in_dtypes(self) -> tuple:
        return tuple(layer.in_dtype for layer in self._layer_summary.values())

    @property
    def out_dtypes(self) -> tuple:
        return tuple(layer.out_dtype for layer in self._layer_summary.values())

    @property
    def param_nums(self) -> tuple[int, ...]:
        return tuple(layer.num_parameters for layer in self._layer_summary.values())

    @property
    def learn_param_nums(self) -> tuple[int, ...]:
        return tuple(layer.num_learnable_parameters for layer in self._layer_summary.values())

    def summarize(self) -> dict[str, ModuleSummary]:
        summary = OrderedDict((name, ModuleSummary(module)) for name, module in self.named_modules)
        model_forward_dummy_input(self._model)
        for layer in summary.values():
            layer.detach_hook()
        return summary

    def _create_table(self, tablefmt: str = "fancy_grid") -> str:
        tab = {
            "Name": self.layer_names,
            "Type": self.layer_types,
            "Params": list(map(convert_human_readable_count, self.param_nums)),
            "Learn Params": list(map(convert_human_readable_count, self.learn_param_nums)),
        }

        if hasattr(self._model, "get_dummy_input"):
            tab["In sizes"] = multiline_format_size(self.in_sizes)
            tab["In dtype"] = multiline_format_dtype(self.in_dtypes)
            tab["Out sizes"] = multiline_format_size(self.out_sizes)
            tab["Out dtype"] = multiline_format_dtype(self.out_dtypes)

        return tabulate(tab, headers="keys", tablefmt=tablefmt, showindex="always")

    def __str__(self) -> str:
        summary = self._create_table()

        learnable_parameters = sum(p.numel() for p in self._model.parameters() if p.requires_grad)
        total_parameters = sum(p.numel() for p in self._model.parameters())
        non_learnable_parameters = total_parameters - learnable_parameters

        summary += (
            f"\n - {convert_human_readable_count(learnable_parameters):<{10}} Learnable params\n"
        )
        summary += (
            f" - {convert_human_readable_count(non_learnable_parameters):<{10}} "
            f"Non-learnable params\n"
        )
        summary += f" - {convert_human_readable_count(total_parameters):<{10}} Total params\n"
        return summary


def model_forward_dummy_input(model: Module) -> None:
    r"""Runs the example input through each layer to get input- and output sizes
    and data types.

    Args:
        model (``torch.nn.Module``): Specifies the model. The model
            should have the ``get_dummy_input`` method. This function
            is a noop if the model does not have the
            ``get_dummy_input`` method.
    """
    if not hasattr(model, "get_dummy_input"):
        return
    dummy_input = model.get_dummy_input()
    mode = model.training
    model.eval()
    with torch.no_grad():
        # Let the model hooks collect the input and output shapes and data types.
        if torch.is_tensor(dummy_input):
            model(dummy_input)
        else:
            model(*dummy_input)
    model.train(mode)  # restore mode of module



r"""This module implements some multiplication-like fusion layers."""

__all__ = ["MultiplicationFusion"]

from torch import Tensor
from torch.nn import Module


class MultiplicationFusion(Module):
    r"""Defines a fusion layer that multiplies the inputs.

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.nn import MultiplicationFusion
        >>> module = MultiplicationFusion()
        >>> module
        MultiplicationFusion()
        >>> x1 = torch.tensor([[2, 3, 4], [5, 6, 7]], dtype=torch.float, requires_grad=True)
        >>> x2 = torch.tensor([[12, 13, 14], [15, 16, 17]], dtype=torch.float, requires_grad=True)
        >>> out = module(x1, x2)
        >>> out
        tensor([[ 24.,  39.,  56.],
                [ 75.,  96., 119.]], grad_fn=<MulBackward0>)
        >>> out.mean().backward()
    """

    def forward(self, *inputs: Tensor) -> Tensor:
        r"""Multiplies the list or tuple of inputs.

        Args:
            *inputs (list or tuple of tensors): Specifies the list or
                tuple of tensors to multiply. The shape of the tensors
                should be the same. By default, this layer expects
                that each input is a ``torch.Tensor`` of shape
                ``(batch size, feature size)``. But it can also work
                if the inputs have a shape
                ``(sequence length, batch size, feature size)`` or
                similar shapes.

        Returns:
            ``torch.Tensor`` with the same shape that the input
                tensor: The fused tensor.
        """
        if len(inputs) == 0:
            raise ValueError(
                f"{type(self).__qualname__} must have at least one tensor in the input"
            )
        output = inputs[0]
        for xi in inputs[1:]:
            output = output.mul(xi)
        return output



r"""This module implements a data source for MNIST dataset."""

__all__ = ["MnistDataSource"]

from pathlib import Path
from typing import Union

from torch.utils.data import Dataset

from gravitorch import constants as ct
from gravitorch.creators.dataloader.base import BaseDataLoaderCreator
from gravitorch.data.datasets import MNISTDataset
from gravitorch.datasources.dataset import DatasetDataSource
from gravitorch.utils.integrations import is_torchvision_available
from gravitorch.utils.path import sanitize_path

if is_torchvision_available():
    from torchvision import transforms
else:
    transforms = None  # pragma: no cover


class MnistDataSource(DatasetDataSource):
    r"""Implements a data source for the MNIST dataset.

    Args:
        path (str, optional): Specifies the path where to save/load
            the MNIST data.
        data_loader_creators (dict): Specifies the data loader
            creators to initialize. Each key indicates a data loader
            creator name. The value can be a ``BaseDataLoaderCreator``
            object, or its configuration, or ``None``. ``None`` means
            a default data loader will be created. Each data loader
            creator takes a ``Dataset`` object as input, so you need
            to specify a dataset with the same name. The expected
            keys are ``'train'`` and ``'eval'``.
    """

    def __init__(
        self,
        path: Union[Path, str],
        data_loader_creators: dict[str, Union[BaseDataLoaderCreator, dict, None]],
    ):
        train_dataset, eval_dataset = create_standard_train_eval_dataset(sanitize_path(path))
        super().__init__(
            datasets={ct.TRAIN: train_dataset, ct.EVAL: eval_dataset},
            data_loader_creators=data_loader_creators,
        )


def create_standard_train_eval_dataset(path: Path) -> tuple[Dataset, Dataset]:
    r"""Creates the standard training and evaluation MNIST datasets.

    Args:
        path (``pathlib.Path``): Specifies the path where to save/load
            MNIST data.

    Returns:
        ``Dataset, Dataset``: The training (first value in the tuple)
            and evaluation (second value in the tuple) datasets.
    """
    transform = transforms.Compose(
        [transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))]
    )
    return (
        MNISTDataset(path.as_posix(), train=True, download=True, transform=transform),
        MNISTDataset(path.as_posix(), train=False, download=True, transform=transform),
    )



__all__ = ["BaseMLP", "AlphaMLP", "BetaMLP", "create_alpha_mlp", "create_beta_mlp"]

from collections import OrderedDict
from collections.abc import Sequence
from typing import Optional

import torch
from objectory import OBJECT_TARGET
from torch import Tensor, nn

from gravitorch import constants as ct
from gravitorch.nn.utils.factory import setup_nn_module
from gravitorch.nn.utils.module_helpers import get_module_device, get_module_name


class BaseMLP(nn.Module):
    r"""Defines a base class to implement a MLP architecture.

    Args:
        input_size (int): Specifies the input size of the MLP network.
        input_name (str, optional): Specifies the name of the input.
            This name is used to identify the input in the batch and
            to export the network. Default: ``'input'``
        output_name (str, optional): Specifies the name of the output.
            This name is used to create the network output dictionary
            and to export the network. Default: ``'prediction'``
    """

    def __init__(
        self,
        input_size: int,
        input_name: str = ct.INPUT,
        output_name: str = ct.PREDICTION,
    ):
        super().__init__()
        self._input_size = input_size
        self._input_name = input_name
        self._output_name = output_name

    def __len__(self) -> int:
        return len(self.layers)

    @property
    def input_size(self) -> int:
        r"""int: The input size of the MLP network."""
        return self._input_size

    def forward(self, inputs: Tensor) -> Tensor:
        r"""Computes the predictions of the MLP network.

        Args:
            inputs (``torch.Tensor`` of shape
                ``(batch_size, input_size)`` and type ``float``):
                Specifies the input of the MLP network.

        Returns:
            ``torch.Tensor`` of shape ``(batch_size, output_size)``
                and type ``float``: The predictions.
        """
        return self.layers(inputs)

    def get_dummy_input(self, batch_size: int = 1) -> tuple[Tensor]:
        r"""Generates a dummy input for the MLP.

        Args:
            batch_size (int, optional): Specifies the batch size to
                use to generate the dummy input. Default: ``1``

        Returns:
            ``tuple[Tensor]``: A tuple with one tensor of shape
                ``(batch_size, input_size)`` and type ``float``.
                The tensor is on the same device that this network.
        """
        return (torch.randn(batch_size, self._input_size, device=get_module_device(self.layers)),)

    def get_input_names(self) -> tuple[str]:
        r"""Gets the tuple of input names for the network.

        The order of the name should be the same that the order in the
        inputs of the forward function.

        Returns:
            tuple: The tuple of input names.
        """
        return (self._input_name,)

    def get_onnx_dynamic_axis(self) -> dict:
        r"""Gets the dynamic axes (e.g. batch size or sequence length) when
        serializing a model to ONNX format.

        See https://pytorch.org/docs/stable/onnx.html#torch.onnx.export
        to have more information on how to create the ``dict``.

        Returns:
            dict: with the dynamic axes of the input/output.
        """
        return {self._input_name: {0: "batch"}, self._output_name: {0: "batch"}}

    def get_output_names(self) -> tuple[str]:
        r"""Gets the tuple of output names for the network.

        The order of the name should be the same that the order in
        the outputs of the forward function.

        Returns:
            tuple: The tuple of output names.
        """
        return (self._output_name,)


class AlphaMLP(BaseMLP):
    r"""Implements a MLP network where the last layer is an activation layer.

    Args:
        input_size (int): Specifies the input size of the MLP network.
        hidden_sizes (sequence): Specifies the hidden sizes of the
            MLP network. The last size is the output size of the MLP.
            This input should have at least one value.
        activation (dict or ``None``, optional): Specifies the
            configuration the activation layer. Default: ``None``
        dropout (float, optional): Specifies the dropout probability
            of an element to be zeroed. Default: ``0.0``
        input_name (str, optional): Specifies the name of the input.
            This name is used to identify the input in the batch and
            to export the network. Default: ``'input'``
        output_name (str, optional): Specifies the name of the output.
            This name is used to create the network output
            dictionary and to export the network.
            Default: ``'prediction'``
    """

    def __init__(
        self,
        input_size: int,
        hidden_sizes: Sequence[int],
        activation: Optional[dict] = None,
        dropout: float = 0.0,
        input_name: str = ct.INPUT,
        output_name: str = ct.PREDICTION,
    ):
        super().__init__(input_size=input_size, input_name=input_name, output_name=output_name)
        self.layers = create_alpha_mlp(
            input_size=input_size,
            hidden_sizes=hidden_sizes,
            activation=activation,
            dropout=dropout,
        )


class BetaMLP(BaseMLP):
    r"""Implements a MLP network where the last layer is a linear layer (without
    activation).

    Args:
        input_size (int): Specifies the input size of the MLP network.
        hidden_sizes (sequence): Specifies the hidden sizes of the
            MLP network. The last size is the output size of the MLP.
            This input should have at least one value.
        activation (dict or ``None``, optional): Specifies the
            configuration the activation layer. Default: ``None``
        dropout (float, optional): Specifies the dropout probability
            of an element to be zeroed. Default: ``0.0``
        input_name (str, optional): Specifies the name of the input.
            This name is used to identify the input in the batch and
            to export the network. Default: ``'input'``
        output_name (str, optional): Specifies the name of the output.
            This name is used to create the network output
            dictionary and to export the network.
            Default: ``'prediction'``
    """

    def __init__(
        self,
        input_size: int,
        hidden_sizes: Sequence[int],
        activation: Optional[dict] = None,
        dropout: float = 0.0,
        input_name: str = ct.INPUT,
        output_name: str = ct.PREDICTION,
    ):
        super().__init__(input_size=input_size, input_name=input_name, output_name=output_name)
        self.layers = create_beta_mlp(
            input_size=input_size,
            hidden_sizes=hidden_sizes,
            activation=activation,
            dropout=dropout,
        )


def create_alpha_mlp(
    input_size: int,
    hidden_sizes: Sequence[int],
    activation: Optional[dict] = None,
    dropout: float = 0.0,
) -> nn.Sequential:
    r"""Creates a MLP network where the last layer is an activation layer.

    Args:
        input_size (int): Specifies the input size of the MLP network.
        hidden_sizes (sequence): Specifies the hidden sizes of the
            MLP network. The last size is the output size of the MLP.
            This input should have at least one value.
        activation (dict or ``None``, optional): Specifies the
            configuration the activation layer. Default: ``None``
        dropout (float, optional): Specifies the dropout probability
            of an element to be zeroed. Default: ``0.0``

    Returns:
        ``torch.nn.Sequential``: The instantiated MLP network.
    """
    activation = activation or {OBJECT_TARGET: "torch.nn.ReLU"}
    sizes = (input_size, *hidden_sizes)
    layers = OrderedDict()
    for i in range(len(sizes) - 1):
        if dropout > 0:
            layers[f"dropout{i + 1}"] = nn.Dropout(dropout)
        layers[f"linear{i + 1}"] = nn.Linear(sizes[i], sizes[i + 1])
        activation_layer = setup_nn_module(activation)
        layers[f"{get_module_name(activation_layer).lower()}{i + 1}"] = activation_layer
    return nn.Sequential(layers)


def create_beta_mlp(
    input_size: int,
    hidden_sizes: Sequence[int],
    activation: Optional[dict] = None,
    dropout: float = 0.0,
) -> nn.Sequential:
    r"""Creates a MLP network where the last layer is a linear layer (without
    activation).

    Args:
        input_size (int): Specifies the input size of the MLP network.
        hidden_sizes (sequence): Specifies the hidden sizes of the
            MLP network. The last size is the output size of the MLP.
            This input should have at least one value.
        activation (dict or ``None``, optional): Specifies the
            configuration the activation layer. Default: ``None``
        dropout (float, optional): Specifies the dropout probability
            of an element to be zeroed. Default: ``0.0``

    Returns:
        ``torch.nn.Sequential``: The instantiated MLP network.
    """
    activation = activation or {OBJECT_TARGET: "torch.nn.ReLU"}
    sizes = (input_size, *hidden_sizes)
    layers = OrderedDict()
    for i in range(len(sizes) - 1):
        if dropout > 0:
            layers[f"dropout{i + 1}"] = nn.Dropout(dropout)
        layers[f"linear{i + 1}"] = nn.Linear(sizes[i], sizes[i + 1])
        activation_layer = setup_nn_module(activation)
        if i < len(sizes) - 2:
            layers[f"{get_module_name(activation_layer).lower()}{i + 1}"] = activation_layer
    return nn.Sequential(layers)



__all__ = [
    "AsinhScalarEncoder",
    "AsinhCosSinScalarEncoder",
    "CosSinScalarEncoder",
    "ScalarEncoderFFN",
]

import math
from typing import Any, Union

import torch
from torch import Tensor
from torch.nn import Module, Parameter

from gravitorch.nn.utils.factory import setup_nn_module
from gravitorch.utils.tensor.misc import to_tensor


class AsinhScalarEncoder(Module):
    r"""Implements a scalar encoder using the inverse hyperbolic sine (arcsinh).

    Args:
        scale (``torch.nn.Tensor`` or list or tuple): Specifies
            the initial scale values.
        learnable (bool, optional): If ``True`` the scale parameters
            are learnable, otherwise they are frozen.
            Default: ``False``
    """

    def __init__(
        self,
        scale: Union[Tensor, list[float], tuple[float, ...]],
        learnable: bool = False,
    ):
        super().__init__()
        self.scale = Parameter(to_tensor(scale), requires_grad=learnable)

    @property
    def input_size(self) -> int:
        r"""int: The input feature size."""
        return 1

    def extra_repr(self) -> str:
        return f"dim={self.scale.shape[0]}, learnable={self.scale.requires_grad}"

    def forward(self, scalar: Tensor):
        r"""Computes a scalar representation.

        Args:
            scalar (``torch.Tensor`` of type float and shape
                ``(*, 1)``): Specifies the scalar values to encode.

        Returns:
            ``torch.Tensor`` of type float and shape
                ``(*, output_size)``: The scalar representation.
        """
        return scalar.mul(self.scale).asinh()

    @classmethod
    def create_rand_scale(
        cls,
        dim: int,
        min_scale: float,
        max_scale: float,
        learnable: bool = False,
    ) -> "AsinhScalarEncoder":
        r"""Creates a `AsinhScalarEncoder`` where the scales are uniformly
        initialized in the specified scale range.

        Args:
            dim (int): Specifies the dimension i.e. the number of
                scales.
            min_scale (float): Specifies the minimum scale.
            max_scale (float): Specifies the maximum scale.
            learnable (bool, optional): If ``True`` the scales are
                learnable, otherwise they are frozen.
                Default: ``False``

        Returns:
            ``AsinhScalarEncoder``: An instantiated
                ``AsinhScalarEncoder`` where the scales are
                uniformly initialized in a scale range.
        """
        if dim < 1:
            raise ValueError(f"dim has to be greater or equal to 1 (received: {dim})")
        if min_scale <= 0:
            raise ValueError(f"min_scale has to be greater than 0 (received: {min_scale})")
        if max_scale < min_scale:
            raise ValueError(
                f"max_scale has to be greater than min_scale {min_scale} "
                f"(received: {max_scale})"
            )
        return cls(
            scale=torch.rand(dim).mul(max_scale - min_scale).add(min_scale),
            learnable=learnable,
        )

    @classmethod
    def create_linspace_scale(
        cls,
        dim: int,
        min_scale: float,
        max_scale: float,
        learnable: bool = False,
    ) -> "AsinhScalarEncoder":
        r"""Creates a `AsinhScalarEncoder`` where the scales are evenly spaced.

        Args:
            dim (int): Specifies the dimension i.e. the number of
                scales.
            min_scale (float): Specifies the minimum scale.
            max_scale (float): Specifies the maximum scale.
            learnable (bool, optional): If ``True`` the scales are
                learnable, otherwise they are frozen.
                Default: ``False``

        Returns:
            ``AsinhScalarEncoder``: An instantiated
                ``AsinhScalarEncoder`` where the scales are
                evenly spaced.
        """
        if dim < 1:
            raise ValueError(f"dim has to be greater or equal to 1 (received: {dim})")
        if min_scale <= 0:
            raise ValueError(f"min_scale has to be greater than 0 (received: {min_scale})")
        if max_scale < min_scale:
            raise ValueError(
                f"max_scale has to be greater than min_scale {min_scale} "
                f"(received: {max_scale})"
            )
        return cls(
            scale=torch.linspace(start=min_scale, end=max_scale, steps=dim),
            learnable=learnable,
        )

    @classmethod
    def create_logspace_scale(
        cls,
        dim: int,
        min_scale: float,
        max_scale: float,
        learnable: bool = False,
    ) -> "AsinhScalarEncoder":
        r"""Creates a `AsinhScalarEncoder`` where the scales are evenly spaced
        in the log space.

        Args:
            dim (int): Specifies the dimension i.e. the number of
                scales.
            min_scale (float): Specifies the minimum scale.
            max_scale (float): Specifies the maximum scale.
            learnable (bool, optional): If ``True`` the scales are
                learnable, otherwise they are frozen.
                Default: ``False``

        Returns:
            ``AsinhScalarEncoder``: An instantiated
                ``AsinhScalarEncoder`` where the scales are
                evenly spaced in the log space.
        """
        if dim < 1:
            raise ValueError(f"dim has to be greater or equal to 1 (received: {dim})")
        if min_scale <= 0:
            raise ValueError(f"min_scale has to be greater than 0 (received: {min_scale})")
        if max_scale < min_scale:
            raise ValueError(
                f"max_scale has to be greater than min_scale {min_scale} "
                f"(received: {max_scale})"
            )
        return cls(
            scale=torch.logspace(start=math.log10(min_scale), end=math.log10(max_scale), steps=dim),
            learnable=learnable,
        )


class CosSinScalarEncoder(Module):
    r"""Implements a frequency/phase-shift scalar encoder where the periodic
    functions are cosine and sine.

    Args:
        frequency (``torch.nn.Tensor`` or list or tuple): Specifies
            the initial frequency values.
        phase_shift (``torch.nn.Tensor`` or list or tuple): Specifies
            the initial phase-shift values.
        learnable (bool, optional): If ``True`` the frequencies and
            phase-shift parameters are learnable, otherwise they are
            frozen. Default: ``False``
    """

    def __init__(
        self,
        frequency: Union[Tensor, list[float], tuple[float, ...]],
        phase_shift: Union[Tensor, list[float], tuple[float, ...]],
        learnable: bool = False,
    ):
        super().__init__()
        frequency = to_tensor(frequency)
        phase_shift = to_tensor(phase_shift)
        if frequency.ndim != 1:
            raise ValueError(
                f"Incorrect number of dimensions for `frequency` (shape={frequency.shape}). "
                f"`frequency` has to be a 1D tensor or equivalent."
            )
        if frequency.shape != phase_shift.shape:
            raise ValueError(
                f"Incorrect shapes. The shape of `frequency` (shape={frequency.shape})"
                f"does not match with `phase_shift` (shape={phase_shift.shape})"
            )
        self.frequency = Parameter(frequency, requires_grad=learnable)
        self.phase_shift = Parameter(phase_shift, requires_grad=learnable)
        self._half_size = int(self.frequency.shape[0] // 2)

    @property
    def input_size(self) -> int:
        r"""int: The input feature size."""
        return 1

    def extra_repr(self) -> str:
        return f"dim={self.frequency.shape[0]}, learnable={self.frequency.requires_grad}"

    def forward(self, scalar: Tensor):
        r"""Computes a scalar representation.

        Args:
            scalar (``torch.Tensor`` of type float and shape
                ``(*, 1)``): Specifies the scalar values to encode.

        Returns:
            ``torch.Tensor`` of type float and shape
                ``(*, output_size)``: The scalar representation.
        """
        features = self.frequency * scalar + self.phase_shift
        return torch.cat(
            (features[..., : self._half_size].sin(), features[..., self._half_size :].cos()),
            dim=-1,
        )

    @classmethod
    def create_rand_frequency(
        cls,
        num_frequencies: int,
        min_frequency: float,
        max_frequency: float,
        learnable: bool = True,
    ) -> "CosSinScalarEncoder":
        r"""Creates a `CosSinScalarEncoder`` where the frequencies are uniformly
        initialized in a frequency range.

        Args:
            num_frequencies (int): Specifies the number of
                frequencies.
            min_frequency (float): Specifies the minimum frequency.
            max_frequency (float): Specifies the maximum frequency.
            learnable (bool, optional): If ``True`` the parameters
                are learnable, otherwise they are frozen.
                Default: ``True``

        Returns:
            ``CosSinScalarEncoder``: An instantiated
                ``CosSinScalarEncoder`` where the frequencies are
                uniformly initialized in a frequency range.
        """
        if num_frequencies < 1:
            raise ValueError(
                f"num_frequencies has to be greater or equal to 1 (received: {num_frequencies})"
            )
        if min_frequency <= 0:
            raise ValueError(f"min_frequency has to be greater than 0 (received: {min_frequency})")
        if max_frequency < min_frequency:
            raise ValueError(
                f"max_frequency has to be greater than min_frequency {min_frequency} "
                f"(received: {max_frequency})"
            )
        return cls(
            frequency=torch.rand(num_frequencies)
            .mul(max_frequency - min_frequency)
            .add(min_frequency)
            .repeat(2),
            phase_shift=torch.zeros(2 * num_frequencies),
            learnable=learnable,
        )

    @classmethod
    def create_rand_value_range(
        cls,
        num_frequencies: int,
        min_abs_value: float,
        max_abs_value: float,
        learnable: bool = True,
    ) -> "CosSinScalarEncoder":
        r"""Creates a `CosSinScalarEncoder`` where the frequencies are uniformly
        initialized for a given value range.

        Args:
            num_frequencies (int): Specifies the number of
                frequencies.
            min_abs_value (float): Specifies the minimum absolute
                value to encode.
            max_abs_value (float): Specifies the maximum absolute
                value to encoder.
            learnable (bool, optional): If ``True`` the parameters
                are learnable, otherwise they are frozen.
                Default: ``True``

        Returns:
            ``CosSinScalarEncoder``: An instantiated
                ``CosSinScalarEncoder`` where the frequencies are
                uniformly initialized for a given value range.
        """
        return cls.create_rand_frequency(
            num_frequencies=num_frequencies,
            min_frequency=1 / max_abs_value,
            max_frequency=1 / min_abs_value,
            learnable=learnable,
        )

    @classmethod
    def create_linspace_frequency(
        cls,
        num_frequencies: int,
        min_frequency: float,
        max_frequency: float,
        learnable: bool = True,
    ) -> "CosSinScalarEncoder":
        r"""Creates a `CosSinScalarEncoder`` where the frequencies are evenly
        spaced in a frequency range.

        Args:
            num_frequencies (int): Specifies the number of
                frequencies.
            min_frequency (float): Specifies the minimum frequency.
            max_frequency (float): Specifies the maximum frequency.
            learnable (bool, optional): If ``True`` the parameters
                are learnable, otherwise they are frozen.
                Default: ``True``

        Returns:
            ``CosSinScalarEncoder``: An instantiated
                ``CosSinScalarEncoder`` where the frequencies are
                evenly spaced in a frequency range.
        """
        if num_frequencies < 1:
            raise ValueError(
                f"num_frequencies has to be greater or equal to 1 (received: {num_frequencies})"
            )
        if min_frequency <= 0:
            raise ValueError(f"min_frequency has to be greater than 0 (received: {min_frequency})")
        if max_frequency < min_frequency:
            raise ValueError(
                f"max_frequency has to be greater than min_frequency {min_frequency} "
                f"(received: {max_frequency})"
            )
        return cls(
            frequency=torch.linspace(
                start=min_frequency, end=max_frequency, steps=num_frequencies
            ).repeat(2),
            phase_shift=torch.zeros(2 * num_frequencies),
            learnable=learnable,
        )

    @classmethod
    def create_linspace_value_range(
        cls,
        num_frequencies: int,
        min_abs_value: float,
        max_abs_value: float,
        learnable: bool = True,
    ) -> "CosSinScalarEncoder":
        r"""Creates a `CosSinScalarEncoder`` where the frequencies are evenly
        spaced given a value range.

        Args:
            num_frequencies (int): Specifies the number of
                frequencies.
            min_abs_value (float): Specifies the minimum absolute
                value to encode.
            max_abs_value (float): Specifies the maximum absolute
                value to encoder.
            learnable (bool, optional): If ``True`` the parameters
                are learnable, otherwise they are frozen.
                Default: ``True``

        Returns:
            ``CosSinScalarEncoder``: An instantiated
                ``CosSinScalarEncoder`` where the frequencies are
                evenly spaced.
        """
        return cls.create_linspace_frequency(
            num_frequencies=num_frequencies,
            min_frequency=1 / max_abs_value,
            max_frequency=1 / min_abs_value,
            learnable=learnable,
        )

    @classmethod
    def create_logspace_frequency(
        cls,
        num_frequencies: int,
        min_frequency: float,
        max_frequency: float,
        learnable: bool = True,
    ) -> "CosSinScalarEncoder":
        r"""Creates a `CosSinScalarEncoder`` where the frequencies are evenly
        spaced in the log space in a frequency range.

        Args:
            num_frequencies (int): Specifies the number of
                frequencies.
            min_frequency (float): Specifies the minimum frequency.
            max_frequency (float): Specifies the maximum frequency.
            learnable (bool, optional): If ``True`` the parameters
                are learnable, otherwise they are frozen.
                Default: ``True``

        Returns:
            ``CosSinScalarEncoder``: An instantiated
                ``CosSinScalarEncoder`` where the frequencies are
                evenly spaced in the log space in a frequency range.
        """
        if num_frequencies < 1:
            raise ValueError(
                f"num_frequencies has to be greater or equal to 1 (received: {num_frequencies})"
            )
        if min_frequency <= 0:
            raise ValueError(f"min_frequency has to be greater than 0 (received: {min_frequency})")
        if max_frequency < min_frequency:
            raise ValueError(
                f"max_frequency has to be greater than min_frequency {min_frequency} "
                f"(received: {max_frequency})"
            )
        return cls(
            frequency=torch.logspace(
                start=math.log10(min_frequency),
                end=math.log10(max_frequency),
                steps=num_frequencies,
            ).repeat(2),
            phase_shift=torch.zeros(2 * num_frequencies),
            learnable=learnable,
        )

    @classmethod
    def create_logspace_value_range(
        cls,
        num_frequencies: int,
        min_abs_value: float,
        max_abs_value: float,
        learnable: bool = True,
    ) -> "CosSinScalarEncoder":
        r"""Creates a `CosSinScalarEncoder`` where the frequencies are evenly
        spaced in the log space given a value range.

        Args:
            num_frequencies (int): Specifies the number of
                frequencies.
            min_abs_value (float): Specifies the minimum absolute
                value to encode.
            max_abs_value (float): Specifies the maximum absolute
                value to encoder.
            learnable (bool, optional): If ``True`` the parameters
                are learnable, otherwise they are frozen.
                Default: ``True``

        Returns:
            ``CosSinScalarEncoder``: An instantiated
                ``CosSinScalarEncoder`` where the frequencies are
                evenly spaced in the log space.
        """
        return cls.create_logspace_frequency(
            num_frequencies=num_frequencies,
            min_frequency=1 / max_abs_value,
            max_frequency=1 / min_abs_value,
            learnable=learnable,
        )


class AsinhCosSinScalarEncoder(CosSinScalarEncoder):
    r"""Extension of ``CosSinScalarEncoder`` with an extra feature:

    the inverse hyperbolic sine (arcsinh) of the input.
    """

    def forward(self, scalar: Tensor):
        r"""Computes a scalar representation.

        Args:
            scalar (``torch.Tensor`` of type float and shape
                ``(d0, d1, ..., dn, 1)``): Specifies the scalar
                values to encode.

        Returns:
            ``torch.Tensor`` of type float and shape
                ``(d0, d1, ..., dn, output_size)``: The scalar
                representation.
        """
        features = self.frequency * scalar + self.phase_shift
        return torch.cat(
            (
                features[..., : self._half_size].sin(),
                features[..., self._half_size :].cos(),
                scalar.asinh(),
            ),
            dim=-1,
        )


class ScalarEncoderFFN(Module):
    r"""Implements a scalar encoder with a feed-forward network.

    Args:
        encoder (``torch.nn.Module`` or dict): Specifies the encoder
            or its configuration.
        ffn (``torch.nn.Module`` or dict): Specifies the feed-forward
            network or its configuration.
    """

    def __init__(self, encoder: Union[Module, dict[str, Any]], ffn: Union[Module, dict[str, Any]]):
        super().__init__()
        self.encoder = setup_nn_module(encoder)
        self.ffn = setup_nn_module(ffn)

    @property
    def input_size(self) -> int:
        r"""int: The input feature size."""
        return self.encoder.input_size

    def forward(self, scalar: Tensor) -> torch.Tensor:
        r"""Computes a scalar representation.

        Args:
            scalar (``torch.Tensor`` of type float and shape
                ``(*, 1)``): Specifies the scalar values to encode.

        Returns:
            ``torch.Tensor`` of type float and shape
                ``(*, output_size)``: The scalar representation.
        """
        return self.ffn(self.encoder(scalar))



__all__ = ["BaseConfusionMatrix", "BinaryConfusionMatrix", "MulticlassConfusionMatrix"]

from collections.abc import Iterable, Sequence
from typing import Optional, Union

import torch
from torch import Tensor

from gravitorch.distributed.ddp import SUM, sync_reduce_
from gravitorch.utils.format import str_add_indent
from gravitorch.utils.meters.exceptions import EmptyMeterError


class BaseConfusionMatrix:
    r"""Defines the base class to implement confusion matrix.

    Args:
        matrix (``torch.Tensor`` of type long and shape
            ``(num_classes, num_classes)``): Specifies the initial
            confusion matrix values. The rows indicate the true
            labels and the columns indicate the predicted labels.
    """

    def __init__(self, matrix: Tensor):
        check_confusion_matrix(matrix)
        self._matrix = matrix
        self._num_predictions = self._compute_num_predictions()

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  num_predictions={self.num_predictions:,}\n"
            f"  shape={self._matrix.shape}\n"
            f"  dtype={self._matrix.dtype}\n"
            f"  {str_add_indent(self._matrix)}\n)"
        )

    def __str__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(num_classes={self.num_classes:,}, "
            f"num_predictions={self.num_predictions:,})"
        )

    @property
    def matrix(self) -> Tensor:
        r"""``torch.Tensor`` of type long and shape ``(num_classes,
        num_classes)``: The confusion matrix values."""
        return self._matrix

    @property
    def num_classes(self) -> int:
        r"""int: The number of classes."""
        return self._matrix.shape[0]

    @property
    def num_predictions(self) -> int:
        r"""int: The number of predictions."""
        return self._num_predictions

    def all_reduce(self) -> None:
        r"""Reduces the meter values across all machines in such a way that all
        get the final result.

        The confusion matrix is reduced by summing all the confusion
        matrices (1 confusion matrix per distributed process).
        """
        sync_reduce_(self._matrix, SUM)
        # It is necessary to recompute the number of predictions because
        # the confusion matrix may have changed
        self._num_predictions = self._compute_num_predictions()

    def get_normalized_matrix(self, normalization: str) -> Tensor:
        r"""Gets the normalized confusion matrix.

        Args:
            normalization (str): Specifies the normalization strategy.
                The supported normalization strategies are:

                    - ``'true'``: normalization over the targets
                        (most commonly used)
                    - ``'pred'``: normalization over the predictions
                    - ``'all'``: normalization over the whole matrix

        Returns:
            ``torch.Tensor`` of type float and shape
                ``(num_classes, num_classes)``: The normalized
                confusion matrix.

        Raises:
            ValueError if the normalization strategy is not supported.
        """
        if normalization == "true":
            # Clamp to avoid division by 0
            return self.matrix / self.matrix.sum(dim=1, keepdim=True).clamp(min=1e-8)
        if normalization == "pred":
            # Clamp to avoid division by 0
            return self.matrix / self.matrix.sum(dim=0, keepdim=True).clamp(min=1e-8)
        if normalization == "all":
            # Clamp to avoid division by 0
            return self.matrix / self.matrix.sum().clamp(min=1e-8)
        raise ValueError(
            f"Incorrect normalization {normalization}. The supported normalization strategies "
            "are `true`, `pred` and `all`"
        )

    def reset(self) -> None:
        r"""Resets the confusion matrix."""
        self._matrix.zero_()
        self._num_predictions = 0

    def update(self, prediction: Tensor, target: Tensor) -> None:
        r"""Updates the confusion matrix with new predictions.

        Args:
            prediction (``torch.Tensor`` of type long and shape
                ``(d0, d1, ..., dn)``): Specifies the predicted labels.
            target (``torch.Tensor`` of type long and shape
                ``(d0, d1, ..., dn)``): Specifies the ground truth
                labels.
        """
        self._matrix += (
            torch.bincount(
                (target.flatten() * self.num_classes + prediction.flatten()).long(),
                minlength=self.num_classes**2,
            )
            .reshape(self.num_classes, self.num_classes)
            .to(device=self._matrix.device)
        )
        self._num_predictions = self._compute_num_predictions()

    def _compute_num_predictions(self) -> int:
        return self._matrix.sum().item()


class BinaryConfusionMatrix(BaseConfusionMatrix):
    r"""Implements a confusion matrix for binary labels.

    Args:
        matrix (``torch.Tensor`` of type long and shape ``(2, 2)``):
            Specifies the initial confusion matrix values.
    """

    def __init__(self, matrix: Optional[Tensor] = None):
        if matrix is None:
            matrix = torch.zeros(2, 2, dtype=torch.long)
        if matrix.shape != (2, 2):
            raise ValueError
        super().__init__(matrix)

    def clone(self) -> "BinaryConfusionMatrix":
        r"""Creates a copy of the current confusion matrix meter.

        Returns:
            ``BinaryConfusionMatrix``: A copy of the current confusion
                matrix meter.
        """
        return BinaryConfusionMatrix(self.matrix.clone())

    def equal(self, other) -> bool:
        r"""Indicates if two confusion matrices are equal or not.

        Args:
            other: Specifies the value to compare.

        Returns:
            bool: ``True`` if the confusion matrices are equal,
                ``False`` otherwise.
        """
        if not isinstance(other, BinaryConfusionMatrix):
            return False
        return self.matrix.equal(other.matrix)

    @classmethod
    def from_predictions(cls, prediction: Tensor, target: Tensor) -> "BinaryConfusionMatrix":
        r"""Creates a confusion matrix given ground truth and predicted labels.

        Args:
            prediction (``torch.Tensor`` of type long and shape
                ``(d0, d1, ..., dn)``): Specifies the predicted labels.
            target (``torch.Tensor`` of type long and shape
                ``(d0, d1, ..., dn)``): Specifies the ground truth
                labels.
        """
        confmat = cls()
        confmat.update(prediction, target)
        return confmat

    ##########################
    #     Transformation     #
    ##########################

    def __add__(self, other):
        return self.add(other)

    def __iadd__(self, other):
        self.add_(other)
        return self

    def __sub__(self, other):
        return self.sub(other)

    def add(self, other: "BinaryConfusionMatrix") -> "BinaryConfusionMatrix":
        r"""Adds a confusion matrix.

        Args:
            other (``BinaryConfusionMatrix``): Specifies the other
                confusion matrix to add.

        Returns:
            ``BinaryConfusionMatrix``: A new confusion matrix
                containing the addition of the two confusion matrices.
        """
        check_op_compatibility_binary(self, other, "add")
        return BinaryConfusionMatrix(self.matrix.add(other.matrix))

    def add_(self, other: "BinaryConfusionMatrix") -> None:
        r"""Adds a confusion matrix.

        In-place version of ``add``.

        Args:
            other (``BinaryConfusionMatrix``): Specifies the other
                confusion matrix to add.
        """
        check_op_compatibility_binary(self, other, "add")
        self.matrix.add_(other.matrix)
        self._num_predictions = self._compute_num_predictions()

    def merge(self, meters: Iterable["BinaryConfusionMatrix"]) -> "BinaryConfusionMatrix":
        r"""Merges several meters with the current meter and returns a new
        meter.

        Args:
            meters (iterable): Specifies the meters to merge to the
                current meter.

        Returns:
            ``BinaryConfusionMatrix``: The merged meter.
        """
        output = self.clone()
        for meter in meters:
            output.add_(meter)
        return output

    def merge_(self, meters: Iterable["BinaryConfusionMatrix"]) -> None:
        r"""Merges several meters into the current meter.

        In-place version of ``merge``.

        Args:
            meters (iterable): Specifies the meters to merge to the
                current meter.
        """
        for meter in meters:
            self.add_(meter)

    def sub(self, other: "BinaryConfusionMatrix") -> "BinaryConfusionMatrix":
        r"""Subtracts a confusion matrix.

        Args:
            other (``BinaryConfusionMatrix``): Specifies the other
                confusion matrix to subtract.

        Returns:
            ``BinaryConfusionMatrix``: A new confusion matrix
                containing the difference of the two confusion
                matrices.
        """
        check_op_compatibility_binary(self, other, "sub")
        return BinaryConfusionMatrix(self.matrix.sub(other.matrix))

    ###################
    #     Metrics     #
    ###################

    @property
    def false_negative(self) -> int:
        r"""int: The false negative i.e. the number of incorrectly
        classified negative examples."""
        return self._matrix[0, 1].item()

    @property
    def false_positive(self) -> int:
        r"""int: The false positive i.e. the number of incorrectly
        classified positive examples."""
        return self._matrix[1, 0].item()

    @property
    def negative(self) -> int:
        r"""int: The number of negative true labels."""
        return self.true_negative + self.false_positive

    @property
    def positive(self) -> int:
        r"""int: The number of positive true labels."""
        return self.true_positive + self.false_negative

    @property
    def predictive_negative(self) -> int:
        r"""int: The number of negative predictions."""
        return self.false_negative + self.true_negative

    @property
    def predictive_positive(self) -> int:
        r"""int: The number of positive predictions."""
        return self.true_positive + self.false_positive

    @property
    def true_negative(self) -> int:
        r"""int: The true negative i.e. the number of correctly
        classified negative examples."""
        return self._matrix[1, 1].item()

    @property
    def true_positive(self) -> int:
        r"""int: The true positive i.e. the number of correctly
        classified positive examples."""
        return self._matrix[0, 0].item()

    def accuracy(self) -> float:
        r"""Computes the accuracy.

        Returns:
            float: The accuracy.

        Raises:
            ``EmptyMeterError`` if the confusion matrix is empty.
        """
        if self.num_predictions == 0:
            raise EmptyMeterError(
                "It is not possible to compute the accuracy because the confusion matrix is empty"
            )
        return float(self.true_positive + self.true_negative) / float(self._num_predictions)

    def balanced_accuracy(self) -> float:
        r"""Computes the balanced accuracy.

        Returns:
            float: The balanced accuracy.

        Raises:
            ``EmptyMeterError`` if the confusion matrix is empty.
        """
        if self.num_predictions == 0:
            raise EmptyMeterError(
                "It is not possible to compute the balanced accuracy because the confusion matrix "
                "is empty"
            )
        return (self.true_negative_rate() + self.true_positive_rate()) / 2

    def f_beta_score(self, beta: float = 1.0) -> float:
        r"""Computes the F-beta score.

        Args:
            beta (float, optional): Specifies the beta value.
                Default: ``1.0``

        Returns:
            float: the F-beta score.

        Raises:
            ``EmptyMeterError`` if the confusion matrix is empty.
        """
        if self.num_predictions == 0:
            raise EmptyMeterError(
                "It is not possible to compute the F-beta score because the confusion matrix "
                "is empty"
            )
        beta2 = beta**2
        return ((1.0 + beta2) * self.true_positive) / (
            (1.0 + beta2) * self.true_positive + beta2 * self.false_negative + self.false_positive
        )

    def false_negative_rate(self) -> float:
        r"""Computes the false negative rate i.e. the miss rate.

        Returns:
            float: The false negative rate.

        Raises:
            ``EmptyMeterError`` if the confusion matrix is empty.
        """
        if self.num_predictions == 0:
            raise EmptyMeterError(
                "It is not possible to compute the false negative rate because the confusion "
                "matrix is empty"
            )
        return float(self.false_negative) / float(self.positive)

    def false_positive_rate(self) -> float:
        r"""Computes the false positive rate i.e. the probability of false
        alarm.

        Returns:
            float: The false positive rate.

        Raises:
            ``EmptyMeterError`` if the confusion matrix is empty.
        """
        if self.num_predictions == 0:
            raise EmptyMeterError(
                "It is not possible to compute the false positive rate because the confusion "
                "matrix is empty"
            )
        return float(self.false_positive) / float(self.negative)

    def jaccard_index(self) -> float:
        r"""Computes the Jaccard index.

        Returns:
            float: The Jaccard index.

        Raises:
            ``EmptyMeterError`` if the confusion matrix is empty.
        """
        if self.num_predictions == 0:
            raise EmptyMeterError(
                "It is not possible to compute the Jaccard index because the confusion "
                "matrix is empty"
            )
        return float(self.true_positive) / float(
            self.true_positive + self.false_negative + self.false_positive
        )

    def precision(self) -> float:
        r"""Computes the precision.

        Returns:
            float: The precision.

        Raises:
            ``EmptyMeterError`` if the confusion matrix is empty.
        """
        if self.num_predictions == 0:
            raise EmptyMeterError(
                "It is not possible to compute the precision because the confusion "
                "matrix is empty"
            )
        return float(self.true_positive) / float(self.predictive_positive)

    def recall(self) -> float:
        r"""Computes the recall i.e. the probability of positive detection.

        Returns:
            float: The recall.

        Raises:
            ``EmptyMeterError`` if the confusion matrix is empty.
        """
        if self.num_predictions == 0:
            raise EmptyMeterError(
                "It is not possible to compute the recall because the confusion " "matrix is empty"
            )
        return float(self.true_positive) / float(self.positive)

    def true_negative_rate(self) -> float:
        r"""Computes the true negative rate.

        Returns:
            float: The true negative rate.

        Raises:
            ``EmptyMeterError`` if the confusion matrix is empty.
        """
        if self.num_predictions == 0:
            raise EmptyMeterError(
                "It is not possible to compute the true negative rate because the confusion "
                "matrix is empty"
            )
        return float(self.true_negative) / float(self.negative)

    def true_positive_rate(self) -> float:
        r"""Computes the true positive rate.

        Returns:
            float: The true positive rate.

        Raises:
            ``EmptyMeterError`` if the confusion matrix is empty.
        """
        if self.num_predictions == 0:
            raise EmptyMeterError(
                "It is not possible to compute the true positive rate because the confusion "
                "matrix is empty"
            )
        return self.recall()

    def compute_all_metrics(
        self,
        betas: Sequence[Union[int, float]] = (1,),
        prefix: str = "",
        suffix: str = "",
    ) -> dict[str, float]:
        r"""Computes all the metrics.

        Args:
            betas (sequence, optional): Specifies the betas used to
                compute the f-beta score. Default: ``(1,)``
            prefix (str, optional): Specifies a prefix for all the
                metrics. Default: ``''``
            suffix (str, optional): Specifies a suffix for all the
                metrics. Default: ``''``

        Returns:
            dict: All the metrics.

        Raises:
            ``EmptyMeterError`` if the confusion matrix is empty.
        """
        if self.num_predictions == 0:
            raise EmptyMeterError(
                "It is not possible to compute the metrics because the confusion matrix is empty"
            )
        metrics = {
            f"{prefix}accuracy{suffix}": self.accuracy(),
            f"{prefix}balanced_accuracy{suffix}": self.balanced_accuracy(),
            f"{prefix}false_negative_rate{suffix}": self.false_negative_rate(),
            f"{prefix}false_positive_rate{suffix}": self.false_positive_rate(),
            f"{prefix}jaccard_index{suffix}": self.jaccard_index(),
            f"{prefix}precision{suffix}": self.precision(),
            f"{prefix}recall{suffix}": self.recall(),
            f"{prefix}true_negative_rate{suffix}": self.true_negative_rate(),
            f"{prefix}true_positive_rate{suffix}": self.true_positive_rate(),
        }
        for beta in betas:
            metrics[f"{prefix}f{beta}_score{suffix}"] = self.f_beta_score(beta)
        return metrics


class MulticlassConfusionMatrix(BaseConfusionMatrix):
    r"""Implements a confusion matrix for multiclass labels."""

    def auto_update(self, prediction: Tensor, target: Tensor) -> None:
        r"""Updates the confusion matrix with new predictions.

        Unlike ``update``, this method will update the number of
        classes if a larger number of classes if found. This method
        allows to use confusion matrix in the setting where the number
        of classes is unknown at the beginning of the process.

        Args:
            prediction (``torch.Tensor`` of type long and shape
                ``(d0, d1, ..., dn)``): Specifies the predicted labels.
            target (``torch.Tensor`` of type long and shape
                ``(d0, d1, ..., dn)``): Specifies the ground truth
                labels.
        """
        # +1 because it is 0-indexed
        num_classes = max(prediction.max().item(), target.max().item()) + 1
        if num_classes > self.num_classes:
            self.resize(num_classes)
        self.update(prediction, target)

    def clone(self) -> "MulticlassConfusionMatrix":
        r"""Creates a copy of the current confusion matrix meter.

        Returns:
            ``MulticlassConfusionMatrix``: A copy of the current confusion matrix meter.
        """
        return MulticlassConfusionMatrix(self.matrix.clone())

    def equal(self, other) -> bool:
        r"""Indicates if two confusion matrices are equal or not.

        Args:
            other: Specifies the value to compare.

        Returns:
            bool: ``True`` if the confusion matrices are equal, ``False`` otherwise.
        """
        if not isinstance(other, MulticlassConfusionMatrix):
            return False
        return self.matrix.equal(other.matrix)

    def resize(self, num_classes: int) -> None:
        r"""Resizes the current confusion matrix to a larger number of classes.

        Args:
            num_classes (int): Specifies the new number of classes.
        """
        if num_classes < self.num_classes:
            raise ValueError(
                f"Incorrect number of classes: {num_classes}. The confusion matrix "
                f"(num_classes={self.num_classes}) can be resized only to a larger number "
                "of classes"
            )
        matrix = self._matrix
        self._matrix = torch.zeros(num_classes, num_classes, dtype=torch.long)
        self._matrix[: matrix.shape[0], : matrix.shape[1]] = matrix

    @classmethod
    def from_num_classes(cls, num_classes: int) -> "MulticlassConfusionMatrix":
        r"""Creates a confusion matrix given the number of classes.

        Args:
            num_classes (int): Specifies the number of classes.

        Returns:
            ``MulticlassConfusionMatrix``: An instantiated confusion
                matrix.
        """
        if num_classes < 1:
            raise ValueError(
                "Incorrect `num_classes`. `num_classes` has to be greater or equal to 1 but "
                f"received {num_classes}"
            )
        return cls(matrix=torch.zeros(num_classes, num_classes, dtype=torch.long))

    @classmethod
    def from_predictions(cls, prediction: Tensor, target: Tensor) -> "MulticlassConfusionMatrix":
        r"""Creates a confusion matrix given ground truth and predicted labels.

        Note: the number of classes is inferred from the maximum
        ground truth and predicted labels.

        Args:
            prediction (``torch.Tensor`` of type long and shape
                ``(d0, d1, ..., dn)``): Specifies the predicted labels.
            target (``torch.Tensor`` of type long and shape
                ``(d0, d1, ..., dn)``): Specifies the ground truth
                labels.

        Returns:
            ``MulticlassConfusionMatrix``: An instantiated confusion matrix.
        """
        # use a fake number of classes. `auto_update` will find the right number of classes
        confmat = cls.from_num_classes(num_classes=1)
        confmat.auto_update(prediction, target)
        return confmat

    ##########################
    #     Transformation     #
    ##########################

    def __add__(self, other):
        return self.add(other)

    def __iadd__(self, other):
        self.add_(other)
        return self

    def __sub__(self, other):
        return self.sub(other)

    def add(self, other: "MulticlassConfusionMatrix") -> "MulticlassConfusionMatrix":
        r"""Adds a confusion matrix.

        Args:
            other (``MulticlassConfusionMatrix``): Specifies the other
                confusion matrix to add.

        Returns:
            ``MulticlassConfusionMatrix``: A new confusion matrix
                containing the addition of the two confusion matrices.
        """
        check_op_compatibility_multiclass(self, other, "add")
        return MulticlassConfusionMatrix(self.matrix.add(other.matrix))

    def add_(self, other: "MulticlassConfusionMatrix") -> None:
        r"""Adds a confusion matrix.

        In-place version of ``add``.

        Args:
            other (``MulticlassConfusionMatrix``): Specifies the other
                confusion matrix to add.
        """
        check_op_compatibility_multiclass(self, other, "add")
        self.matrix.add_(other.matrix)
        self._num_predictions = self._compute_num_predictions()

    def merge(self, meters: Iterable["MulticlassConfusionMatrix"]) -> "MulticlassConfusionMatrix":
        r"""Merges several meters with the current meter and returns a new
        meter.

        Args:
            meters (iterable): Specifies the meters to merge to the
                current meter.

        Returns:
            ``MulticlassConfusionMatrix``: The merged meter.
        """
        output = self.clone()
        for meter in meters:
            output.add_(meter)
        return output

    def merge_(self, meters: Iterable["MulticlassConfusionMatrix"]) -> None:
        r"""Merges several meters into the current meter.

        In-place version of ``merge``.

        Args:
            meters (iterable): Specifies the meters to merge to the
                current meter.
        """
        for meter in meters:
            self.add_(meter)

    def sub(self, other: "MulticlassConfusionMatrix") -> "MulticlassConfusionMatrix":
        r"""Subtracts a confusion matrix.

        Args:
            other (``MulticlassConfusionMatrix``): Specifies the other
                confusion matrix to subtract.

        Returns:
            ``MulticlassConfusionMatrix``: A new confusion matrix
                containing the difference of the two confusion matrices.
        """
        check_op_compatibility_multiclass(self, other, "sub")
        return MulticlassConfusionMatrix(self.matrix.sub(other.matrix))

    ###################
    #     Metrics     #
    ###################

    @property
    def false_negative(self) -> Tensor:
        r"""``torch.Tensor`` of type long and shape ``(num_classes,)``:

        The number of false negative for each class i.e. the elements
        that have been labelled as negative by the model, but they are
        actually positive.
        """
        return self.support - self.true_positive

    @property
    def false_positive(self) -> Tensor:
        r"""``torch.Tensor`` of type long and shape ``(num_classes,)``:

        The number of false positive for each class i.e. the elements
        that have been labelled as positive by the model, but they are
        actually negative.
        """
        return self.matrix.sum(dim=0) - self.true_positive

    @property
    def support(self) -> Tensor:
        r"""``torch.Tensor`` of type long and shape ``(num_classes,)``:

        The support for each class i.e. the number of elements for a
        given class (true label).
        """
        return self.matrix.sum(dim=1)

    @property
    def true_positive(self) -> Tensor:
        r"""``torch.Tensor`` of type long and shape ``(num_classes,)``:

        The number of true positive for each class i.e. the elements
        that have been labelled as positive by the model, and they are
        actually positive.
        """
        return self.matrix.diag()

    def accuracy(self) -> float:
        r"""Computes the accuracy.

        Returns:
            float: The accuracy.

        Raises:
            ``EmptyMeterError`` if the confusion matrix is empty.
        """
        if self.num_predictions == 0:
            raise EmptyMeterError(
                "It is not possible to compute the accuracy because the confusion matrix is empty"
            )
        return float(self.true_positive.sum().item()) / float(self._num_predictions)

    def balanced_accuracy(self) -> float:
        r"""Computes the balanced accuracy.

        Returns:
            float: The balanced accuracy.

        Raises:
            ``EmptyMeterError`` if the confusion matrix is empty.
        """
        if self.num_predictions == 0:
            raise EmptyMeterError(
                "It is not possible to compute the balanced accuracy because the confusion "
                "matrix is empty"
            )
        return self.recall().mean().item()

    def f_beta_score(self, beta: float = 1.0) -> Tensor:
        r"""Computes the F-beta score for each class.

        Args:
            beta (float, optional): Specifies the beta value. Default: ``1.0``

        Returns:
            ``torch.Tensor`` of type float and shape
                ``(num_classes,)``: The F-beta score for each class.

        Raises:
            ``EmptyMeterError`` if the confusion matrix is empty.
        """
        if self.num_predictions == 0:
            raise EmptyMeterError(
                "It is not possible to compute the F-beta score because the confusion matrix "
                "is empty"
            )
        beta2 = beta**2
        return (self.true_positive.mul(1.0 + beta2)) / (
            self.true_positive.mul(1.0 + beta2)
            + self.false_negative.mul(beta2)
            + self.false_positive
        )

    def macro_f_beta_score(self, beta: float = 1.0) -> float:
        r"""Computes the macro (a.k.a. unweighted mean) F-beta score.

        Args:
            beta (float, optional): Specifies the beta value. Default: ``1.0``

        Returns:
            float: The macro F-beta score.

        Raises:
            ``EmptyMeterError`` if the confusion matrix is empty.
        """
        return self.f_beta_score(beta).mean().item()

    def micro_f_beta_score(self, beta: float = 1.0) -> float:
        r"""Computes the micro F-beta score.

        Args:
            beta (float, optional): Specifies the beta value.
                Default: ``1.0``

        Returns:
            float: The micro F-beta score.

        Raises:
            ``EmptyMeterError`` if the confusion matrix is empty.
        """
        if self.num_predictions == 0:
            raise EmptyMeterError(
                "It is not possible to compute the micro F-beta score because the confusion "
                "matrix is empty"
            )
        beta2 = beta**2
        return (
            (self.true_positive.sum().mul(1.0 + beta2))
            / (
                self.true_positive.sum().mul(1.0 + beta2)
                + self.false_negative.sum().mul(beta2)
                + self.false_positive.sum()
            )
        ).item()

    def weighted_f_beta_score(self, beta: float = 1.0) -> float:
        r"""Computes the weighted mean F-beta score.

        Args:
            beta (float, optional): Specifies the beta value.
                Default: ``1.0``

        Returns:
            float: The weighted mean F-beta score.

        Raises:
            ``EmptyMeterError`` if the confusion matrix is empty.
        """
        return self.f_beta_score(beta).mul(self.support).sum().item() / float(self._num_predictions)

    def precision(self) -> Tensor:
        r"""Computes the precision for each class.

        Returns:
            ``torch.Tensor`` of type float and shape
                ``(num_classes,)``: The precision for each class.

        Raises:
            ``EmptyMeterError`` if the confusion matrix is empty.
        """
        if self.num_predictions == 0:
            raise EmptyMeterError(
                "It is not possible to compute the precision because the confusion matrix is empty"
            )
        return self.true_positive.float().div(self.matrix.sum(dim=0).clamp(min=1e-8))

    def macro_precision(self) -> float:
        r"""Computes the macro (a.k.a. unweighted mean) precision.

        Returns:
            float: The macro precision.

        Raises:
            ``EmptyMeterError`` if the confusion matrix is empty.
        """
        return self.precision().mean().item()

    def micro_precision(self) -> float:
        r"""Computes the micro precision.

        Returns:
            float: The micro precision.

        Raises:
            ``EmptyMeterError`` if the confusion matrix is empty.
        """
        if self.num_predictions == 0:
            raise EmptyMeterError(
                "It is not possible to compute the micro precision because the confusion "
                "matrix is empty"
            )
        return (
            self.true_positive.sum()
            .div(self.true_positive.sum().add(self.false_positive.sum()))
            .item()
        )

    def weighted_precision(self) -> float:
        r"""Computes the weighted mean (a.k.a. unweighted mean) precision.

        Returns:
            float: The weighted mean precision.

        Raises:
            ``EmptyMeterError`` if the confusion matrix is empty.
        """
        return self.precision().mul(self.support).sum().item() / float(self._num_predictions)

    def recall(self) -> Tensor:
        r"""Computes the recall for each class.

        Returns:
            ``torch.Tensor`` of type float and shape
                ``(num_classes,)``: The recall for each class.

        Raises:
            ``EmptyMeterError`` if the confusion matrix is empty.
        """
        if self.num_predictions == 0:
            raise EmptyMeterError(
                "It is not possible to compute the recall because the confusion matrix is empty"
            )
        return self.true_positive.float().div(self.support.clamp(min=1e-8))

    def macro_recall(self) -> float:
        r"""Computes the macro (a.k.a. unweighted mean) recall.

        Returns:
            float: The macro recall.

        Raises:
            ``EmptyMeterError`` if the confusion matrix is empty.
        """
        return self.recall().mean().item()

    def micro_recall(self) -> float:
        r"""Computes the micro recall.

        Returns:
            float: The micro recall.

        Raises:
            ``EmptyMeterError`` if the confusion matrix is empty.
        """
        if self.num_predictions == 0:
            raise EmptyMeterError(
                "It is not possible to compute the micro recall because the confusion matrix is empty"
            )
        return (
            self.true_positive.sum()
            .div(self.true_positive.sum().add(self.false_negative.sum()))
            .item()
        )

    def weighted_recall(self) -> float:
        r"""Computes the weighted mean (a.k.a. unweighted mean) recall.

        Returns:
            float: The weighted mean recall.

        Raises:
            ``EmptyMeterError`` if the confusion matrix is empty.
        """
        return self.recall().mul(self.support).sum().item() / float(self._num_predictions)

    def compute_per_class_metrics(
        self,
        betas: Sequence[float] = (1,),
        prefix: str = "",
        suffix: str = "",
    ) -> dict[str, Tensor]:
        r"""Computes all the per-class metrics.

        Args:
            betas (sequence, optional): Specifies the betas used to
                compute the f-beta score. Default: ``(1,)``
            prefix (str, optional): Specifies a prefix for all the
                metrics. Default: ``''``
            suffix (str, optional): Specifies a suffix for all the
                metrics. Default: ``''``

        Returns:
            dict: All the per-class metrics.

        Raises:
            ``EmptyMeterError`` if the confusion matrix is empty.
        """
        if self.num_predictions == 0:
            raise EmptyMeterError(
                "It is not possible to compute the metrics because the confusion matrix is empty"
            )

        metrics = {
            f"{prefix}precision{suffix}": self.precision(),
            f"{prefix}recall{suffix}": self.recall(),
        }
        for beta in betas:
            metrics[f"{prefix}f{beta}_score{suffix}"] = self.f_beta_score(beta)
        return metrics

    def compute_macro_metrics(
        self,
        betas: Sequence[float] = (1,),
        prefix: str = "",
        suffix: str = "",
    ) -> dict[str, float]:
        r"""Computes all the "macro" metrics.

        Args:
            betas (sequence, optional): Specifies the betas used to
                compute the f-beta score. Default: ``(1,)``
            prefix (str, optional): Specifies a prefix for all the
                metrics. Default: ``''``
            suffix (str, optional): Specifies a suffix for all the
                metrics. Default: ``''``

        Returns:
            dict: All the "macro" metrics.

        Raises:
            ``EmptyMeterError`` if the confusion matrix is empty.
        """
        if self.num_predictions == 0:
            raise EmptyMeterError(
                "It is not possible to compute the `macro` metrics because the confusion "
                "matrix is empty"
            )
        metrics = {
            f"{prefix}macro_precision{suffix}": self.macro_precision(),
            f"{prefix}macro_recall{suffix}": self.macro_recall(),
        }
        for beta in betas:
            metrics[f"{prefix}macro_f{beta}_score{suffix}"] = self.macro_f_beta_score(beta)
        return metrics

    def compute_micro_metrics(
        self,
        betas: Sequence[float] = (1,),
        prefix: str = "",
        suffix: str = "",
    ) -> dict[str, float]:
        r"""Computes all the "micro" metrics.

        Args:
            betas (sequence, optional): Specifies the betas used to
                compute the f-beta score. Default: ``(1,)``
            prefix (str, optional): Specifies a prefix for all the
                metrics. Default: ``''``
            suffix (str, optional): Specifies a suffix for all the
                metrics. Default: ``''``

        Returns:
            dict: All the "micro" metrics.

        Raises:
            ``EmptyMeterError`` if the confusion matrix is empty.
        """
        if self.num_predictions == 0:
            raise EmptyMeterError(
                "It is not possible to compute the `micro` metrics because the confusion "
                "matrix is empty"
            )
        metrics = {
            f"{prefix}micro_precision{suffix}": self.micro_precision(),
            f"{prefix}micro_recall{suffix}": self.micro_recall(),
        }
        for beta in betas:
            metrics[f"{prefix}micro_f{beta}_score{suffix}"] = self.micro_f_beta_score(beta)
        return metrics

    def compute_weighted_metrics(
        self,
        betas: Sequence[float] = (1,),
        prefix: str = "",
        suffix: str = "",
    ) -> dict[str, float]:
        r"""Computes all the "weighted" metrics.

        Args:
            betas (sequence, optional): Specifies the betas used to
                compute the f-beta score. Default: ``(1,)``
            prefix (str, optional): Specifies a prefix for all the
                metrics. Default: ``''``
            suffix (str, optional): Specifies a suffix for all the
                metrics. Default: ``''``

        Returns:
            dict: All the "weighted" metrics.

        Raises:
            ``EmptyMeterError`` if the confusion matrix is empty.
        """
        if self.num_predictions == 0:
            raise EmptyMeterError(
                "It is not possible to compute the `weighted` metrics because the confusion "
                "matrix is empty"
            )
        metrics = {
            f"{prefix}weighted_precision{suffix}": self.weighted_precision(),
            f"{prefix}weighted_recall{suffix}": self.weighted_recall(),
        }
        for beta in betas:
            metrics[f"{prefix}weighted_f{beta}_score{suffix}"] = self.weighted_f_beta_score(beta)
        return metrics

    def compute_scalar_metrics(
        self,
        betas: Sequence[float] = (1,),
        prefix: str = "",
        suffix: str = "",
    ) -> dict[str, float]:
        r"""Computes all the scalar metrics.

        Args:
            betas (sequence, optional): Specifies the betas used to
                compute the f-beta score. Default: ``(1,)``
            prefix (str, optional): Specifies a prefix for all the
                metrics. Default: ``''``
            suffix (str, optional): Specifies a suffix for all the
                metrics. Default: ``''``

        Returns:
            dict: All the scalar metrics.

        Raises:
            ``EmptyMeterError`` if the confusion matrix is empty.
        """
        if self.num_predictions == 0:
            raise EmptyMeterError(
                "It is not possible to compute the metrics because the confusion matrix is empty"
            )
        metrics = {
            f"{prefix}accuracy{suffix}": self.accuracy(),
            f"{prefix}balanced_accuracy{suffix}": self.balanced_accuracy(),
        }
        metrics.update(self.compute_macro_metrics(betas, prefix, suffix))
        metrics.update(self.compute_micro_metrics(betas, prefix, suffix))
        metrics.update(self.compute_weighted_metrics(betas, prefix, suffix))
        return metrics


def check_confusion_matrix(matrix: Tensor) -> None:
    r"""Checks if the input matrix is a valid confusion matrix.

    Args:
        matrix (``torch.Tensor``): Specifies the matrix to check.
    """
    if matrix.ndim != 2:
        raise ValueError(
            "Incorrect matrix dimensions. The matrix must have 2 dimensions but "
            f"received {matrix.ndim} dimensions"
        )
    if matrix.shape[0] != matrix.shape[1]:
        raise ValueError(
            "Incorrect matrix shape. The matrix must be a squared matrix but "
            f"received {matrix.shape}"
        )
    if matrix.dtype != torch.long:
        raise ValueError(
            "Incorrect matrix data type. The matrix data type must be long but "
            f"received {matrix.dtype}"
        )
    if not torch.all(matrix >= 0):
        raise ValueError(
            "Incorrect matrix values. The matrix values must be greater or equal to 0 but "
            f"received:\n{matrix}"
        )


def check_op_compatibility_binary(
    current: BinaryConfusionMatrix, other: BinaryConfusionMatrix, op_name: str
) -> None:
    r"""Checks if the confusion matrices for binary labels are compatible.

    Args:
        current (``BinaryConfusionMatrix``): Specifies the current confusion matrix for binary labels.
        other (``BinaryConfusionMatrix``): Specifies the other confusion matrix for binary labels.
        op_name (str): Specifies the operation name.

    Raises:
        TypeError if the other matrix type is not compatible.
    """
    if not isinstance(other, BinaryConfusionMatrix):
        raise TypeError(
            f"Incorrect type {type(other)}. No implementation available to `{op_name}` "
            f"{type(current)} with {type(other)}"
        )


def check_op_compatibility_multiclass(
    current: MulticlassConfusionMatrix, other: MulticlassConfusionMatrix, op_name: str
) -> None:
    r"""Checks if the confusion matrices for multiclass labels are compatible.

    Args:
        current (``MulticlassConfusionMatrix``): Specifies the current
            confusion matrix for multiclass labels.
        other (``MulticlassConfusionMatrix``): Specifies the other
            confusion matrix for multiclass labels.
        op_name (str): Specifies the operation name.

    Raises:
        TypeError if the other matrix type is not compatible.
        ValueError if the matrix shapes are different.
    """
    if not isinstance(other, MulticlassConfusionMatrix):
        raise TypeError(
            f"Incorrect type {type(other)}. No implementation available to `{op_name}` "
            f"{type(current)} with {type(other)}"
        )
    if current.matrix.shape != other.matrix.shape:
        raise ValueError(
            f"Incorrect shape: received {other.matrix.shape} but expect {current.matrix.shape}"
        )



__all__ = ["get_best_values", "get_last_values"]

from collections.abc import Mapping
from typing import Any

from gravitorch.utils.history.base import BaseHistory, EmptyHistoryError


def get_best_values(
    histories: Mapping[str, BaseHistory], prefix: str = "", suffix: str = ""
) -> dict[str, Any]:
    """Gets the best value of each history.

    This function ignores the empty histories and non-comparable
    histories.

    Args:
        histories (``Mapping``): Specifies the histories and their
            keys.
        prefix (str, optional): Specifies the prefix used to create
            the dict of best values. The goal of this prefix is to
            generate a name which is different from the history name
            to avoid confusion. By default, the returned dict uses
            the same name as the history. Default: ``''``
        suffix (str, optional): Specifies the suffix used to create
            the dict of best values. The goal of this suffix is to
            generate a name which is different from the history name
            to avoid confusion. By default, the returned dict uses
            the same name as the history. Default: ``''``

    Returns:
        dict: The dict with the best value of each history.

    Example:

    .. code-block:: python

        >>> from gravitorch.utils.history import (
        ...     MinScalarHistory,
        ...     MaxScalarHistory,
        ...     get_best_values,
        ... )
        >>> history1 = MinScalarHistory('loss')
        >>> history1.add_value(1.9)
        >>> history1.add_value(1.2)
        >>> history2 = MaxScalarHistory('accuracy')
        >>> history2.add_value(42)
        >>> history2.add_value(35)
        >>> get_best_values({'loss': history1, 'accuracy': history2})
        {'loss': 1.2, 'accuracy': 42}
        >>> get_best_values({'loss': history1, 'accuracy': history2}, prefix='best/')
        {'best/loss': 1.2, 'best/accuracy': 42}
        >>> get_best_values({'loss': history1, 'accuracy': history2}, suffix='/best')
        {'loss/best': 1.2, 'accuracy/best': 42}
    """
    values = {}
    for key, history in histories.items():
        if history.is_comparable():
            try:
                values[f"{prefix}{key}{suffix}"] = history.get_best_value()
            except EmptyHistoryError:
                pass
    return values


def get_last_values(
    histories: Mapping[str, BaseHistory], prefix: str = "", suffix: str = ""
) -> dict[str, Any]:
    """Gets the last value of each history.

    This function ignores the empty histories.

    Args:
        histories (``Mapping``): Specifies the histories and their
            keys.
        prefix (str, optional): Specifies the prefix used to create
            the dict of best values. The goal of this prefix is to
            generate a name which is different from the hisotry name
            to avoid confusion. By default, the returned dict uses
            the same name as the history. Default: ``''``
        suffix (str, optional): Specifies the suffix used to create
            the dict of best values. The goal of this suffix is to
            generate a name which is different from the history name
            to avoid confusion. By default, the returned dict uses
            the same name as the history. Default: ``''``

    Returns:
        dict: The dict with the best value of each history.

    Example:

    .. code-block:: python

        >>> from gravitorch.utils.history import (
        ...     MinScalarHistory,
        ...     MaxScalarHistory,
        ...     get_best_values,
        ... )
        >>> history1 = MinScalarHistory('loss')
        >>> history1.add_value(1.9)
        >>> history1.add_value(1.2)
        >>> history2 = MaxScalarHistory('accuracy')
        >>> history2.add_value(42)
        >>> history2.add_value(35)
        >>> get_best_values({'loss': history1, 'accuracy': history2})
        {'loss': 1.2, 'accuracy': 42}
        >>> get_best_values({'loss': history1, 'accuracy': history2}, prefix='best/')
        {'best/loss': 1.2, 'best/accuracy': 42}
        >>> get_best_values({'loss': history1, 'accuracy': history2}, suffix='/best')
        {'loss/best': 1.2, 'accuracy/best': 42}
    """
    values = {}
    for key, history in histories.items():
        try:
            values[f"{prefix}{key}{suffix}"] = history.get_last_value()
        except EmptyHistoryError:
            pass
    return values



__all__ = ["ExtremaTensorMeter", "MeanTensorMeter", "TensorMeter", "TensorMeter2"]

from collections.abc import Iterable
from typing import Any, Optional, Union

import torch
from coola import objects_are_equal
from torch import Tensor

from gravitorch.distributed.ddp import MAX, MIN, SUM, sync_reduce
from gravitorch.utils.format import to_pretty_dict_str
from gravitorch.utils.meters.exceptions import EmptyMeterError
from gravitorch.utils.tensor import scalable_quantile
from gravitorch.utils.tensor.flatted import LazyFlattedTensor


class MeanTensorMeter:
    r"""Implements a meter to compute the mean value of ``torch.Tensor``s.

    The mean value is updated by keeping local variables ``total``
    and ``count``. ``count`` tracks the number of values, and
    ``total`` tracks the sum of the values.
    This meter has a constant space complexity.

    Args:
        count (int, optional): Specifies the initial count value.
            Default: ``0``
        total (float, optional): Specifies the initial total value.
            Default: ``0.0``

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.utils.meters import TensorMeter
        >>> meter = MeanTensorMeter()
        >>> meter.update(torch.arange(6))
        >>> meter.update(torch.tensor([4.0, 1.0]))
        >>> meter.mean()  # or meter.average()
        2.5
        >>> meter.sum()
        20.0
        >>> meter.count
        8
    """

    def __init__(self, count: int = 0, total: Union[int, float] = 0):
        self._count = int(count)
        self._total = total

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(count={self._count:,}, total={self._total})"

    @property
    def count(self) -> int:
        r"""int: The number of predictions in the meter."""
        return self._count

    @property
    def total(self) -> Union[int, float]:
        r"""int or float: The total sum value in the meter."""
        return self._total

    def reset(self) -> None:
        r"""Reset the meter."""
        self._count = 0
        self._total = 0

    def update(self, tensor: Tensor) -> None:
        r"""Updates the meter given a new tensor.

        Args:
            tensor (``torch.Tensor``): Specifies the new tensor to add
                to the meter.
        """
        self._total += tensor.sum().item()
        self._count += tensor.numel()

    def average(self) -> float:
        r"""Computes the average value.

        Returns:
            float: The average value.

        Raises:
            ``EmptyMeterError`` if the meter is empty.
        """
        return self.mean()

    def mean(self) -> float:
        r"""Gets the mean value.

        Returns:
            float: The mean value.

        Raises:
            ``EmptyMeterError`` if the meter is empty.
        """
        if not self._count:
            raise EmptyMeterError("The tensor meter is empty")
        return float(self._total) / float(self._count)

    def sum(self) -> Union[int, float]:
        r"""Gets the sum of all the values.

        Returns:
            int or float: The sum of all the values.

        Raises:
            ``EmptyMeterError`` if the meter is empty.
        """
        if not self._count:
            raise EmptyMeterError("The meter is empty")
        return self._total

    def all_reduce(self) -> "MeanTensorMeter":
        r"""Reduces the meter values across all machines in such a way that all
        get the final result.

        The sum value is reduced by summing all the sum values (1 sum
        value per distributed process). The count value is reduced by
        summing all the count values (1 count value per distributed
        process).

        In a non-distributed setting, this method returns a copy of
        the current meter.

        Returns:
            ``MeanTensorMeter``: The reduced meter.

        Example usage:

        .. code-block:: python

            >>> import torch
            >>> from gravitorch.utils.meters import MeanTensorMeter
            >>> meter = MeanTensorMeter()
            >>> meter.update(torch.arange(6))
            >>> reduced_meter = meter.all_reduce()
        """
        return MeanTensorMeter(
            count=sync_reduce(self._count, SUM), total=sync_reduce(self._total, SUM)
        )

    def clone(self) -> "MeanTensorMeter":
        r"""Creates a copy of the current meter.

        Returns:
            ``MeanTensorMeter``: A copy of the current meter.
        """
        return MeanTensorMeter(count=self._count, total=self._total)

    def equal(self, other) -> bool:
        r"""Indicates if two meters are equal or not.

        Args:
            other: Specifies the value to compare.

        Returns:
            bool: ``True`` if the meters are equal,
                ``False`` otherwise.
        """
        if not isinstance(other, MeanTensorMeter):
            return False
        return self.state_dict() == other.state_dict()

    def merge(self, meters: Iterable["MeanTensorMeter"]) -> "MeanTensorMeter":
        r"""Merges several meters with the current meter and returns a new
        meter.

        Args:
            meters (iterable): Specifies the meters to merge to the
                current meter.

        Returns:
            ``MeanTensorMeter``: The merged meter.
        """
        count, total = self.count, self.total
        for meter in meters:
            count += meter.count
            total += meter.total
        return MeanTensorMeter(total=total, count=count)

    def merge_(self, meters: Iterable["MeanTensorMeter"]) -> None:
        r"""Merges several meters into the current meter.

        In-place version of ``merge``.

        Args:
            meters (iterable): Specifies the meters to merge to the
                current meter.
        """
        for meter in meters:
            self._count += meter.count
            self._total += meter.total

    def load_state_dict(self, state_dict: dict) -> None:
        r"""Loads a state to the history tracker.

        Args:
            state_dict (dict): Specifies a dictionary containing state
                keys with values.
        """
        self._count = state_dict["count"]
        self._total = state_dict["total"]

    def state_dict(self) -> dict[str, Union[int, float]]:
        r"""Returns a dictionary containing state values.

        Returns:
            dict: The state values in a dict.
        """
        return {"count": self._count, "total": self._total}


class ExtremaTensorMeter:
    r"""Implements a meter to compute the minimum and maximum values of
    ``torch.Tensor``s.

    The mean value is updated by keeping local variables ``min_value``
    and ``max_value``. ``min_value`` tracks the minimum value, and
    ``max_value`` tracks the maximum value.
    This meter has a constant space complexity.

    Args:
        count (int, optional): Specifies the initial count value.
            Default: ``0``
        min_value (int, optional): Specifies the initial minimum
            value. Default: ``inf``
        max_value (int, optional): Specifies the initial maximum
            value. Default: ``-inf``

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.utils.meters import ExtremaTensorMeter
        >>> meter = ExtremaTensorMeter()
        >>> meter.update(torch.arange(6))
        >>> meter.update(torch.tensor([4.0, 1.0]))
        >>> meter.max()
        5
        >>> meter.min()
        0
    """

    def __init__(
        self, count: int = 0, min_value: float = float("inf"), max_value: float = float("-inf")
    ):
        self._count = int(count)
        self._min_value = float(min_value)
        self._max_value = float(max_value)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(count={self._count:,}, "
            f"min_value={self._min_value}, max_value={self._max_value})"
        )

    @property
    def count(self) -> int:
        r"""int: The number of predictions in the meter."""
        return self._count

    def reset(self) -> None:
        r"""Reset the meter."""
        self._count = 0
        self._max_value = float("-inf")
        self._min_value = float("inf")

    def update(self, tensor: Tensor) -> None:
        r"""Updates the meter given a new tensor.

        Args:
            tensor (``torch.Tensor``): Specifies the new tensor to add
                to the meter.
        """
        min_value, max_value = torch.aminmax(tensor)
        self._max_value = max(self._max_value, max_value.item())
        self._min_value = min(self._min_value, min_value.item())
        self._count += tensor.numel()

    def max(self) -> float:
        r"""Gets the max value.

        Returns:
            float: The max value.

        Raises:
            ``EmptyMeterError`` if the meter is empty.
        """
        if not self._count:
            raise EmptyMeterError("The meter is empty")
        return float(self._max_value)

    def min(self) -> float:
        r"""Gets the min value.

        Returns:
            float: The min value.

        Raises:
            ``EmptyMeterError`` if the meter is empty.
        """
        if not self._count:
            raise EmptyMeterError("The meter is empty")
        return float(self._min_value)

    def all_reduce(self) -> "ExtremaTensorMeter":
        r"""Reduces the meter values across all machines in such a way that all
        get the final result.

        The maximum value is reduced by computing the maximum between
        the maximum values (1 maximum value per distributed process).
        The minimum value is reduced by computing the minimum between
        the minimum values (1 minimum value per distributed process).

        Returns:
            ``TensorMeter``: The reduced meter.

        Example usage:

        .. code-block:: python

            >>> import torch
            >>> from gravitorch.utils.meters import ExtremaTensorMeter
            >>> meter = ExtremaTensorMeter()
            >>> meter.update(torch.arange(6))
            >>> reduced_meter = meter.all_reduce()
        """
        return ExtremaTensorMeter(
            count=sync_reduce(self._count, SUM),
            min_value=sync_reduce(self._min_value, MIN),
            max_value=sync_reduce(self._max_value, MAX),
        )

    def clone(self) -> "ExtremaTensorMeter":
        r"""Creates a copy of the current meter.

        Returns:
            ``TensorMeter``: A copy of the current meter.
        """
        return ExtremaTensorMeter(
            count=self._count, min_value=self._min_value, max_value=self._max_value
        )

    def equal(self, other) -> bool:
        r"""Indicates if two meters are equal or not.

        Args:
            other: Specifies the value to compare.

        Returns:
            bool: ``True`` if the meters are equal,
                ``False`` otherwise.
        """
        if not isinstance(other, ExtremaTensorMeter):
            return False
        return self.state_dict() == other.state_dict()

    def merge(self, meters: Iterable["ExtremaTensorMeter"]) -> "ExtremaTensorMeter":
        r"""Merges several meters with the current meter and returns a new
        meter.

        Args:
            meters (iterable): Specifies the meters to merge to the
                current meter.

        Returns:
            ``ExtremaTensorMeter``: The merged meter.
        """
        count, min_value, max_value = self._count, self._min_value, self._max_value
        for meter in meters:
            count += meter.count
            min_value = min(min_value, meter._min_value)
            max_value = max(max_value, meter._max_value)
        return ExtremaTensorMeter(count=count, min_value=min_value, max_value=max_value)

    def merge_(self, meters: Iterable["ExtremaTensorMeter"]) -> None:
        r"""Merges several meters into the current meter.

        In-place version of ``merge``.

        Args:
            meters (iterable): Specifies the meters to merge to the
                current meter.
        """
        for meter in meters:
            self._count += meter.count
            self._min_value = min(self._min_value, meter._min_value)
            self._max_value = max(self._max_value, meter._max_value)

    def load_state_dict(self, state_dict: dict[str, Any]) -> None:
        r"""Loads a state to the history tracker.

        Args:
            state_dict (dict): Specifies a dictionary containing state
                keys with values.
        """
        self._count = state_dict["count"]
        self._min_value = state_dict["min_value"]
        self._max_value = state_dict["max_value"]

    def state_dict(self) -> dict[str, Any]:
        r"""Returns a dictionary containing state values.

        Returns:
            dict: The state values in a dict.
        """
        return {"count": self._count, "max_value": self._max_value, "min_value": self._min_value}


class TensorMeter:
    r"""Defines a class to compute and store the sum, average, maximum and
    minimum values of ``torch.Tensor``s.

    This meter has a constant space complexity.

    Args:
        count (int, optional): Specifies the initial count value.
            Default: ``0``
        total (float, optional): Specifies the initial sum value.
            Default: ``0.0``
        min_value (int, optional): Specifies the initial minimum
            value. Default: ``inf``
        max_value (int, optional): Specifies the initial maximum
            value. Default: ``-inf``

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.utils.meters import TensorMeter
        >>> meter = TensorMeter()
        >>> meter.update(torch.arange(6))
        >>> meter.update(torch.tensor([4.0, 1.0]))
        >>> meter.average()  # or meter.mean()
        2.5
        >>> meter.max()
        5
        >>> meter.min()
        0
        >>> meter.sum()
        20.0
        >>> meter.count
        8
    """

    def __init__(
        self,
        count: int = 0,
        total: float = 0.0,
        min_value: float = float("inf"),
        max_value: float = float("-inf"),
    ):
        self._count = int(count)
        self._total = float(total)
        self._min_value = float(min_value)
        self._max_value = float(max_value)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(count={self._count:,}, total={self._total}, "
            f"min_value={self._min_value}, max_value={self._max_value})"
        )

    def __str__(self) -> str:
        count = self.count
        stats = to_pretty_dict_str(
            {
                "count": f"{count:,}",
                "sum": self.sum() if count else "N/A (empty)",
                "average": self.average() if count else "N/A (empty)",
                "min": self.min() if count else "N/A (empty)",
                "max": self.max() if count else "N/A (empty)",
            },
            indent=2,
        )
        return f"{self.__class__.__qualname__}\n{stats}"

    @property
    def count(self) -> int:
        r"""int: The number of predictions in the meter."""
        return self._count

    @property
    def total(self) -> Union[int, float]:
        r"""int or float: The total sum value in the meter."""
        return self._total

    def all_reduce(self) -> "TensorMeter":
        r"""Reduces the meter values across all machines in such a way that all
        get the final result.

        The sum value is reduced by summing all the sum values (1 sum
        value per distributed process). The count value is reduced by
        summing all the count values (1 count value per distributed
        process). The maximum value is reduced by computing the
        maximum between the maximum values (1 maximum value per
        distributed process). The minimum value is reduced by
        computing the minimum between the minimum values (1 minimum
        value per distributed process).

        Returns:
            ``TensorMeter``: The reduced meter.

        Example usage:

        .. code-block:: python

            >>> import torch
            >>> from gravitorch.utils.meters import TensorMeter
            >>> meter = TensorMeter()
            >>> meter.update(torch.arange(6))
            >>> reduced_meter = meter.all_reduce()
        """
        return TensorMeter(
            count=sync_reduce(self._count, SUM),
            total=sync_reduce(self._total, SUM),
            min_value=sync_reduce(self._min_value, MIN),
            max_value=sync_reduce(self._max_value, MAX),
        )

    def clone(self) -> "TensorMeter":
        r"""Creates a copy of the current meter.

        Returns:
            ``TensorMeter``: A copy of the current meter.
        """
        return TensorMeter(
            count=self._count,
            total=self._total,
            min_value=self._min_value,
            max_value=self._max_value,
        )

    def equal(self, other) -> bool:
        r"""Indicates if two meters are equal or not.

        Args:
            other: Specifies the value to compare.

        Returns:
            bool: ``True`` if the meters are equal,
                ``False`` otherwise.
        """
        if not isinstance(other, TensorMeter):
            return False
        return self.state_dict() == other.state_dict()

    def merge(self, meters: Iterable["TensorMeter"]) -> "TensorMeter":
        r"""Merges several meters with the current meter and returns a new
        meter.

        Args:
            meters (iterable): Specifies the meters to merge to the
                current meter.

        Returns:
            ``TensorMeter``: The merged meter.
        """
        count, total = self._count, self._total
        min_value, max_value = self._min_value, self._max_value
        for meter in meters:
            count += meter.count
            total += meter.total
            min_value = min(min_value, meter._min_value)
            max_value = max(max_value, meter._max_value)
        return TensorMeter(total=total, count=count, min_value=min_value, max_value=max_value)

    def merge_(self, meters: Iterable["TensorMeter"]) -> None:
        r"""Merges several meters into the current meter.

        In-place version of ``merge``.

        Args:
            meters (iterable): Specifies the meters to merge to the
                current meter.
        """
        for meter in meters:
            self._count += meter.count
            self._total += meter.total
            self._min_value = min(self._min_value, meter._min_value)
            self._max_value = max(self._max_value, meter._max_value)

    def load_state_dict(self, state_dict: dict[str, Any]) -> None:
        r"""Loads a state to the history tracker.

        Args:
            state_dict (dict): Specifies a dictionary containing state
                keys with values.
        """
        self._count = state_dict["count"]
        self._max_value = state_dict["max_value"]
        self._min_value = state_dict["min_value"]
        self._total = state_dict["total"]

    def state_dict(self) -> dict[str, Any]:
        r"""Returns a dictionary containing state values.

        Returns:
            dict: The state values in a dict.
        """
        return {
            "count": self._count,
            "max_value": self._max_value,
            "min_value": self._min_value,
            "total": self._total,
        }

    def reset(self) -> None:
        r"""Reset the meter."""
        self._count = 0
        self._max_value = float("-inf")
        self._min_value = float("inf")
        self._total = 0.0

    def update(self, tensor: Tensor) -> None:
        r"""Updates the meter given a new tensor.

        Args:
            tensor (``torch.Tensor``): Specifies the new tensor to add
                to the meter.
        """
        min_value, max_value = torch.aminmax(tensor)
        self._max_value = max(self._max_value, max_value.item())
        self._min_value = min(self._min_value, min_value.item())
        self._total += tensor.sum().item()
        self._count += tensor.numel()

    def average(self) -> float:
        r"""Computes the average value.

        Returns:
            float: The average value.

        Raises:
            ``EmptyMeterError`` if the meter is empty.
        """
        if not self._count:
            raise EmptyMeterError("The tensor meter is empty")
        return self._total / float(self._count)

    def max(self) -> float:
        r"""Gets the max value.

        Returns:
            float: The max value.

        Raises:
            ``EmptyMeterError`` if the meter is empty.
        """
        if not self._count:
            raise EmptyMeterError("The meter is empty")
        return float(self._max_value)

    def mean(self) -> float:
        r"""Gets the mean value.

        Returns:
            float: The mean value.

        Raises:
            ``EmptyMeterError`` if the meter is empty.
        """
        if not self._count:
            raise EmptyMeterError("The tensor meter is empty")
        return self._total / float(self._count)

    def min(self) -> float:
        r"""Gets the min value.

        Returns:
            float: The min value.

        Raises:
            ``EmptyMeterError`` if the meter is empty.
        """
        if not self._count:
            raise EmptyMeterError("The meter is empty")
        return float(self._min_value)

    def sum(self) -> float:
        r"""Gets the sum of all the values.

        Returns:
            float: The sum of all the values.

        Raises:
            ``EmptyMeterError`` if the meter is empty.
        """
        if not self._count:
            raise EmptyMeterError("The meter is empty")
        return float(self._total)


class TensorMeter2:
    r"""Implements a meter to compute some stats on ``torch.Tensor``s.

    This meter has a linear space complexity.

    Args:
        values (``torch.Tensor`` or ``None``, optional): Specifies the
            initial values. The tensor is flattened if necessary.
            ``None`` means no initial values. Default: ``None``

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.utils.meters import TensorMeter2
        >>> meter = TensorMeter2()
        >>> meter.update(torch.arange(6))
        >>> meter.update(torch.tensor([4.0, 1.0]))
        >>> meter.count
        8
        >>> meter.average()  # or meter.mean()
        2.5
        >>> meter.max()
        5.0
        >>> meter.min()
        0.0
        >>> meter.sum()
        20.0
        >>> meter.median()
        2.0
        >>> meter.quantile(torch.tensor([0.1, 0.5]))
        tensor([0.7000, 2.5000])
        >>> meter.std()
        1.7728105783462524
    """

    def __init__(self, values: Optional[Tensor] = None):
        self._values = LazyFlattedTensor(values)
        self._count = self._values.numel()

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(count={self._count:,})"

    @property
    def count(self) -> int:
        r"""int: The number of predictions in the meter."""
        return self._count

    def reset(self) -> None:
        r"""Reset the meter."""
        self._count = 0
        self._values.clear()

    def update(self, tensor: Tensor) -> None:
        r"""Updates the meter given a new tensor.

        Args:
            tensor (``torch.Tensor``): Specifies the new tensor to add
                to the meter.
        """
        self._values.update(tensor.detach())
        self._count += tensor.numel()

    def average(self) -> float:
        r"""Computes the average value.

        Returns:
            float: The average value.

        Raises:
            ``EmptyMeterError`` if the meter is empty.
        """
        return self.mean()

    def max(self) -> Union[int, float]:
        r"""Gets the max value.

        Returns:
            int or float: The max value.

        Raises:
            ``EmptyMeterError`` if the meter is empty.
        """
        if not self._count:
            raise EmptyMeterError("The meter is empty")
        return self._values.values().max().item()

    def mean(self) -> float:
        r"""Gets the mean value.

        Returns:
            float: The mean value.

        Raises:
            ``EmptyMeterError`` if the meter is empty.
        """
        if not self._count:
            raise EmptyMeterError("The tensor meter is empty")
        return self._values.values().float().mean().item()

    def median(self) -> float:
        r"""Gets the median value.

        Returns:
            float: The median value.

        Raises:
            ``EmptyMeterError`` if the meter is empty.
        """
        if not self._count:
            raise EmptyMeterError("The meter is empty")
        return self._values.values().median().item()

    def min(self) -> Union[int, float]:
        r"""Gets the min value.

        Returns:
            int or float: The min value.

        Raises:
            ``EmptyMeterError`` if the meter is empty.
        """
        if not self._count:
            raise EmptyMeterError("The meter is empty")
        return self._values.values().min().item()

    def quantile(self, q: Tensor, method: str = "linear") -> Tensor:
        r"""Computes the ``q``-th quantiles.

        Args:
            q (``torch.Tensor`` of type float and shape
                ``(num_q_values,)``): Specifies the ``q``-values in
                the range ``[0, 1]``.
            method (str, optional): Specifies the interpolation
                method to use when the desired quantile lies between
                two data points. Can be ``'linear'``, ``'lower'``,
                ``'higher'``, ``'midpoint'`` and ``'nearest'``.
                Default: ``'linear'``.

        Returns:
            ``torch.Tensor`` of shape  ``(num_q_values,)``: The
                ``q``-th quantiles.

        Raises:
            ``EmptyMeterError`` if the meter is empty.
        """
        if not self._count:
            raise EmptyMeterError("The meter is empty")
        return scalable_quantile(self._values.values().float(), q=q, method=method)

    def std(self) -> float:
        r"""Gets the standard deviation value.

        Returns:
            float: The standard deviation value.

        Raises:
            ``EmptyMeterError`` if the meter is empty .
        """
        if not self._count:
            raise EmptyMeterError("The meter is empty")
        return self._values.values().float().std(dim=0).item()

    def sum(self) -> Union[int, float]:
        r"""Gets the sum of all the values.

        Returns:
            float: The sum of all the values.

        Raises:
            ``EmptyMeterError`` if the meter is empty.
        """
        if not self._count:
            raise EmptyMeterError("The meter is empty")
        return self._values.values().sum().item()

    def all_reduce(self) -> "TensorMeter2":
        r"""Reduces the meter values across all machines in such a way that all
        get the final result.

        Returns:
            ``TensorMeter2``: The reduced meter.

        Example usage:

        .. code-block:: python

            >>> import torch
            >>> from gravitorch.utils.meters import TensorMeter2
            >>> meter = TensorMeter2()
            >>> meter.update(torch.arange(6))
            >>> reduced_meter = meter.all_reduce()
        """
        return TensorMeter2(self._values.all_reduce().values())

    def clone(self) -> "TensorMeter2":
        r"""Creates a copy of the current meter.

        Returns:
            ``TensorMeter2``: A copy of the current meter.
        """
        return TensorMeter2(self._values.clone().values())

    def equal(self, other) -> bool:
        r"""Indicates if two meters are equal or not.

        Args:
            other: Specifies the value to compare.

        Returns:
            bool: ``True`` if the meters are equal,
                ``False`` otherwise.
        """
        if not isinstance(other, TensorMeter2):
            return False
        return objects_are_equal(self.state_dict(), other.state_dict())

    def merge(self, meters: Iterable["TensorMeter2"]) -> "TensorMeter2":
        r"""Merges several meters with the current meter and returns a new
        meter.

        Args:
            meters (iterable): Specifies the meters to merge to the
                current meter.

        Returns:
            ``TensorMeter2``: The merged meter.
        """
        values = self._values.clone()
        for meter in meters:
            values.update(meter._values.values())
        return TensorMeter2(values.values())

    def merge_(self, meters: Iterable["TensorMeter2"]) -> None:
        r"""Merges several meters into the current meter.

        In-place version of ``merge``.

        Args:
            meters (iterable): Specifies the meters to merge to the
                current meter.
        """
        for meter in meters:
            self._values.update(meter._values.values())

    def load_state_dict(self, state_dict: dict[str, Any]) -> None:
        r"""Loads a state to the history tracker.

        Args:
            state_dict (dict): Specifies a dictionary containing state
                keys with values.
        """
        self._values = LazyFlattedTensor(state_dict["values"])

    def state_dict(self) -> dict[str, Tensor]:
        r"""Returns a dictionary containing state values.

        Returns:
            dict: The state values in a dict.
        """
        return {"values": self._values.values()}



__all__ = ["absolute_error", "absolute_relative_error", "symmetric_absolute_relative_error"]

from torch import Tensor


def absolute_error(prediction: Tensor, target: Tensor) -> Tensor:
    r"""Computes the absolute error between the predictions and targets.

    Args:
        prediction (``torch.Tensor``): Specifies the predictions.
        target (``torch.Tensor`` with same shape and data type as
            ``prediction``): Specifies the target tensor.

    Returns:
        ``torch.Tensor`` with same shape and data type as
            the inputs: The absolute error.

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.nn.functional import absolute_error
        >>> absolute_error(torch.eye(2), torch.ones(2, 2))
        tensor([[0., 1.],
                [1., 0.]])
    """
    return target.sub(prediction).abs()


def absolute_relative_error(prediction: Tensor, target: Tensor, eps: float = 1e-8) -> Tensor:
    r"""Computes the absolute relative error between the predictions and
    targets.

    Args:
        prediction (``torch.Tensor``): Specifies the predictions.
        target (``torch.Tensor`` with same shape and data type as
            ``prediction``): Specifies the target tensor.
        eps (float, optional): Specifies an arbitrary small strictly
            positive number to avoid undefined results when the target
            is zero. Default: ``1e-8``

    Returns:
        ``torch.Tensor`` with same shape and data type as
            the inputs: The absolute relative error.

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.nn.functional import absolute_relative_error
        >>> absolute_relative_error(torch.eye(2), torch.ones(2, 2))
        tensor([[0., 1.],
                [1., 0.]])
    """
    return target.sub(prediction).div(target.abs().clamp(min=eps)).abs()


def symmetric_absolute_relative_error(
    prediction: Tensor, target: Tensor, eps: float = 1e-8
) -> Tensor:
    r"""Computes the symmetric absolute relative error between the predictions
    and targets.

    Args:
        prediction (``torch.Tensor``): Specifies the predictions.
        target (``torch.Tensor`` with same shape and data type as
            ``prediction``): Specifies the target tensor.
        eps (float, optional): Specifies an arbitrary small strictly
            positive number to avoid undefined results when the target
            is zero. Default: ``1e-8``

    Returns:
        ``torch.Tensor`` with same shape and data type as
            the inputs: The symmetric absolute relative error.

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.nn.functional import symmetric_absolute_relative_error
        >>> symmetric_absolute_relative_error(torch.eye(2), torch.ones(2, 2))
        tensor([[0., 1.],
                [1., 0.]])
    """
    return (
        target.sub(prediction).div(target.abs().add(prediction.abs()).mul(0.5).clamp(min=eps)).abs()
    )



__all__ = ["AutoDevicePlacement"]

from gravitorch import distributed as dist
from gravitorch.utils.device_placement.manual import ManualDevicePlacement


class AutoDevicePlacement(ManualDevicePlacement):
    r"""Implements a device placement class that automatically find the "best"
    device to use.

    It uses a cuda device if cuda is available, otherwise it uses a cpu
    device.
    """

    def __init__(self):
        super().__init__(dist.device())



__all__ = ["BasePartitioner", "setup_partitioner"]

import logging
from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import Generic, Optional, TypeVar, Union

from objectory import AbstractFactory

from gravitorch.engines.base import BaseEngine
from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)

T = TypeVar("T")


class BasePartitioner(Generic[T], ABC, metaclass=AbstractFactory):
    r"""Defines the base class to implement a partitioner."""

    @abstractmethod
    def partition(
        self, items: Sequence[T], engine: Optional[BaseEngine] = None
    ) -> list[Sequence[T]]:
        r"""Creates data.

        Args:
            items: Specifies the sequence to partition.
            engine (``BaseEngine`` or ``None``): Specifies an engine.
                Default: ``None``

        Return:
            list: The list of partitions.
        """


def setup_partitioner(partitioner: Union[BasePartitioner, dict]) -> BasePartitioner:
    r"""Sets up a partitioner.

    The partitioner is instantiated from its configuration by using
    the ``BasePartitioner`` factory function.

    Args:
        partitioner (``BasePartitioner`` or dict): Specifies the
            partitioner or its configuration.

    Returns:
        ``BasePartitioner``: The instantiated partitioner.
    """
    if isinstance(partitioner, dict):
        logger.info(
            "Initializing a partitioner from its configuration... "
            f"{str_target_object(partitioner)}"
        )
        partitioner = BasePartitioner.factory(**partitioner)
    return partitioner



__all__ = ["create_sequential"]

from collections.abc import Sequence
from typing import Union

from torch.nn import Module, Sequential

from gravitorch.nn.utils.factory import setup_nn_module


def create_sequential(modules: Sequence[Union[Module, dict]]) -> Sequential:
    r"""Creates a ``torch.nn.Sequential`` from a sequence of modules.

    Args:
        modules (sequence): Specifies the sequence of modules or their
            configuration.

    Returns:
        ``torch.nn.Sequential``: The instantiated
            ``torch.nn.Sequential`` module.
    """
    return Sequential(*[setup_nn_module(module) for module in modules])



r"""This module implements a no-operation experiment tracker."""

__all__ = ["NoOpExpTracker"]

import logging
import shutil
import tempfile
from pathlib import Path
from typing import Any, Optional, Union

from gravitorch import constants as ct
from gravitorch.utils.artifacts.base import BaseArtifact
from gravitorch.utils.exp_trackers.base import (
    BaseBasicExpTracker,
    NotActivatedExpTrackerError,
)
from gravitorch.utils.exp_trackers.steps import Step
from gravitorch.utils.integrations import is_matplotlib_available, is_pillow_available
from gravitorch.utils.path import sanitize_path

if is_matplotlib_available():
    from matplotlib.pyplot import Figure
else:
    Figure = "matplotlib.pyplot.Figure"  # pragma: no cover

if is_pillow_available():
    from PIL.Image import Image
else:
    Image = "PIL.Image.Image"  # pragma: no cover

logger = logging.getLogger(__name__)


class NoOpExpTracker(BaseBasicExpTracker):
    r"""Implements a no-operation experiment tracker.

    This class is equivalent to disable the experiment tracker without
    changing the current implementation. You should not use this class
    if you want to track the experiment artifacts. This class creates
    the paths, so you can write files like for the other experiment
    trackers.

    Args:
        experiment_path (``pathlib.Path`` or str or ``None``,
            optional): Specifies the path where to write the
            experiment logs. If ``None``, a temporary directory is
            used. The temporary directory will be removed at the end
            of the experiment. Default: ``None``
    """

    def __init__(self, experiment_path: Union[Path, str, None] = None):
        self._experiment_path = sanitize_path(experiment_path) if experiment_path else None
        # Flag to indicate if the tracker is activated or not
        self._is_activated = False
        # Flag to indicate if the experiment directory should be removed at the end of the run
        self._remove_after_run = False

        # The following directories are defined when the start method is called
        self._checkpoint_path: Optional[Path] = None
        self._artifact_path: Optional[Path] = None

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(experiment_path={self._experiment_path}, "
            f"is_activated={self._is_activated})"
        )

    @property
    def artifact_path(self) -> Path:
        if not self._is_activated:
            raise NotActivatedExpTrackerError(
                "It is not possible to get `artifact_path` because the experiment tracker "
                "is not activated"
            )
        return self._artifact_path

    @property
    def checkpoint_path(self) -> Path:
        if not self._is_activated:
            raise NotActivatedExpTrackerError(
                "It is not possible to get `checkpoint_path` because the experiment tracker "
                "is not activated"
            )
        return self._checkpoint_path

    @property
    def experiment_id(self) -> str:
        if not self._is_activated:
            raise NotActivatedExpTrackerError(
                "It is not possible to get `experiment_id` because the experiment tracker "
                "is not activated"
            )
        return "fakeid0123"

    def _start(self) -> None:
        self._is_activated = True
        if self._experiment_path is None:
            self._experiment_path = sanitize_path(tempfile.mkdtemp())
            self._remove_after_run = True

        logger.info(
            f"Create a noop experiment tracker and store results at {self._experiment_path}"
        )
        self._artifact_path = self._experiment_path.joinpath(ct.ARTIFACT_FOLDER_NAME)
        self._checkpoint_path = self._experiment_path.joinpath(ct.CHECKPOINT_FOLDER_NAME)

    def _flush(self, upload_checkpoints: bool = True) -> None:
        pass  # Do nothing because it is a no-operation experiment tracker.

    def _end(self) -> None:
        if self._remove_after_run:
            # Remove the temporary directory
            shutil.rmtree(self._experiment_path)
            self._experiment_path = None

        self._artifact_path = None
        self._checkpoint_path = None
        self._is_activated = False

    def is_activated(self) -> bool:
        return self._is_activated

    def _is_resumed(self) -> bool:
        return False  # Always none because this experiment tracker cannot be resumed.

    def _add_tag(self, name: str, value: Any) -> None:
        pass  # Do nothing because it is a no-operation experiment tracker.

    def _add_tags(self, tags: dict[str, Any]) -> None:
        pass  # Do nothing because it is a no-operation experiment tracker.

    def _create_artifact(self, artifact: BaseArtifact) -> None:
        pass  # Do nothing because it is a no-operation experiment tracker.

    def _log_best_metric(self, key: str, value: Union[int, float]) -> None:
        pass  # Do nothing because it is a no-operation experiment tracker.

    def _log_best_metrics(self, metrics: dict[str, Union[int, float]]) -> None:
        pass  # Do nothing because it is a no-operation experiment tracker.

    def _log_figure(self, key: str, figure: Figure, step: Optional[Step] = None) -> None:
        pass  # Do nothing because it is a no-operation experiment tracker.

    def _log_figures(self, figures: dict[str, Figure], step: Optional[Step] = None) -> None:
        pass  # Do nothing because it is a no-operation experiment tracker.

    def _log_hyper_parameter(self, key: str, value: Any) -> None:
        pass  # Do nothing because it is a no-operation experiment tracker.

    def _log_hyper_parameters(self, params: dict[str, Any]) -> None:
        pass  # Do nothing because it is a no-operation experiment tracker.

    def _log_image(self, key: str, image: Image, step: Optional[Step] = None) -> None:
        pass  # Do nothing because it is a no-operation experiment tracker.

    def _log_images(self, images: dict[str, Image], step: Optional[Step] = None) -> None:
        pass  # Do nothing because it is a no-operation experiment tracker.

    def _log_metric(self, key: str, value: Union[int, float], step: Optional[Step] = None) -> None:
        pass  # Do nothing because it is a no-operation experiment tracker.

    def _log_metrics(
        self, metrics: dict[str, Union[int, float]], step: Optional[Step] = None
    ) -> None:
        pass  # Do nothing because it is a no-operation experiment tracker.

    def _upload_checkpoints(self) -> None:
        pass  # Do nothing because it is a no-operation experiment tracker.



__all__ = ["SequentialPartitioner"]

from collections.abc import Sequence
from typing import Optional, TypeVar

from gravitorch.data.partitioners.base import BasePartitioner
from gravitorch.distributed import comm as dist
from gravitorch.engines import BaseEngine

T = TypeVar("T")


class SequentialPartitioner(BasePartitioner[T]):
    r"""Implements a partitioner that select the items in the partition
    sequentially.

    If ``partition_size=1``, the partitioner returns the first item
    for the first epoch, the second items for the second epoch, etc.

    Args:
        partition_size (int, optional): Specifies the partition size
            i.e. the number of items in each partition.
            Default: ``1``
    """

    def __init__(self, partition_size: int = 1):
        self._partition_size = int(partition_size)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(partition_size={self.partition_size:,})"

    @property
    def partition_size(self) -> int:
        r"""int: The partition size."""
        return self._partition_size

    def partition(
        self, items: Sequence[T], engine: Optional[BaseEngine] = None
    ) -> list[Sequence[T]]:
        if not items:
            return [[]]
        world_size = dist.get_world_size()
        epoch = engine.epoch if engine is not None else 0
        num_items = len(items)
        return [
            [
                items[j % num_items]
                for j in range(
                    epoch * world_size * self._partition_size + i * self._partition_size,
                    epoch * world_size * self._partition_size + (i + 1) * self._partition_size,
                )
            ]
            for i in range(world_size)
        ]



__all__ = ["FlattenModule", "MulticlassFlattenModule"]

from typing import Any, Union

import torch
from torch import Tensor
from torch.nn import Module

from gravitorch.nn.utils import setup_nn_module


class FlattenModule(Module):
    r"""Implements a wrapper to flat the inputs of a ``torch.nn.Module``.

    Args:
        module (``torch.nn.Module`` or dict): Specifies the module or
            its configuration.
    """

    def __init__(self, module: Union[Module, dict]):
        super().__init__()
        self.module = setup_nn_module(module)

    def forward(self, *args, **kwargs) -> Any:
        r"""Computes the forward pass of the module.

        All the inputs are converted to a 1d tensor.

        Args:
            *args: Same as ``module``
            **kwargs: Same as ``module``

        Returns:
            Same as ``module``
        """
        return self.module(
            *[torch.flatten(arg) for arg in args],
            **{key: torch.flatten(value) for key, value in kwargs.items()},
        )


class MulticlassFlattenModule(FlattenModule):
    r"""Implements a wrapper to flat the multiclass inputs of a
    ``torch.nn.Module``."""

    def forward(self, prediction: Tensor, target: Tensor) -> Any:
        r"""Computes the forward pass of the module.

        Args:
            prediction (``torch.Tensor`` of shape
                ``(batch_size, d1, d2, ..., dn, C)``): Specifies the
                predictions. This tensor will be reshaped to
                ``(batch_size * d1 * d2 * ... * dn, C)``.
            target (``torch.Tensor`` of shape
                ``(batch_size, d1, d2, ..., dn)``): Specifies the
                targets. This tensor will be reshaped to
                ``(batch_size * d1 * d2 * ... * dn,)``.

        Returns:
            Same as ``module``
        """
        return self.module(prediction.view(-1, prediction.shape[-1]), torch.flatten(target))



from __future__ import annotations

__all__ = ["BaseLRSchedulerCreator"]

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from objectory import AbstractFactory
from torch.optim import Optimizer

from gravitorch.lr_schedulers.base import LRSchedulerType

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine


class BaseLRSchedulerCreator(ABC, metaclass=AbstractFactory):
    r"""Defines the base class to create a learning rate (LR) scheduler.

    Note that it is not the unique approach to create a LR scheduler.
    Feel free to use other approaches if this approach does not fit your
    needs.
    """

    @abstractmethod
    def create(self, engine: BaseEngine, optimizer: Optimizer | None) -> LRSchedulerType | None:
        r"""Creates an optimizer.

        This method is responsible to register the event handlers
        associated to the LR scheduler. In particular, it should
        register the event to call the ``step`` method of the LR
        scheduler. If the optimizer is ``None``, this function should
        return ``None``  because it does not make sense to define a LR
        scheduler without an optimizer.

        Args:
            engine (``gravitorch.engines.BaseEngine``): Specifies an
                engine.
            optimizer (``torch.nn.Optimizer``): Specifies the
                optimizer.

        Returns:
            ``LRSchedulerType`` or ``None``: The created LR scheduler
                or ``None`` if there is no LR scheduler to create.
        """



r"""This package contains the implementation of some profilers."""

__all__ = ["BaseProfiler", "NoOpProfiler", "PyTorchProfiler", "setup_profiler"]

from gravitorch.utils.profilers.base import BaseProfiler
from gravitorch.utils.profilers.noop import NoOpProfiler
from gravitorch.utils.profilers.pytorch import PyTorchProfiler
from gravitorch.utils.profilers.utils import setup_profiler



r"""This module implements a base class to compute a descriptive summary of
sequences."""

__all__ = ["BaseSequenceDataSummary"]

from abc import abstractmethod
from typing import TypeVar

from gravitorch.utils.data_summary.base import BaseDataSummary
from gravitorch.utils.data_summary.float_value import FloatDataSummary

T = TypeVar("T")


class BaseSequenceDataSummary(BaseDataSummary[T]):
    r"""Implements a base class to compute a descriptive summary of sequences.

    A child class has to implement the ``_get_sequence_length`` method.

    Args:
        value_summary (``BaseDataSummary``): Specifies the summary
            object used to compute a descriptive summary of the
            values in the sequences.
    """

    def __init__(self, value_summary: BaseDataSummary[T]):
        self._value_summary = value_summary
        self._length_summary = FloatDataSummary()

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  value_summary={self._value_summary},\n"
            f"  length_summary={self._length_summary}\n"
            ")"
        )

    def add(self, data: T) -> None:
        r"""Adds new data to the summary.

        Args:
            data: Specifies the data to add to the summary.
        """
        self._value_summary.add(data)
        self._length_summary.add(self._get_sequence_length(data))

    def reset(self) -> None:
        r"""Resets the data summary."""
        self._value_summary.reset()
        self._length_summary.reset()

    def summary(self) -> dict:
        r"""Gets a descriptive summary of the data.

        Returns:
            dict: The data descriptive summary. The dictionary has two
                keys: ``'length'`` and ``'value'``. The key
                ``'length'`` contains some information about the
                sequence length. The key ``'value'`` contains some
                information about the values in the sequence.

        Raises:
            ``EmptyDataSummaryError`` is the data summary is empty.
        """
        return {
            "value": self._value_summary.summary(),
            "length": self._length_summary.summary(),
        }

    @abstractmethod
    def _get_sequence_length(self, data: T) -> int:
        r"""Gets the sequence length of the data.

        Args:
            data : Specifies the input sequence.

        Returns:
            int: The sequence length.
        """



r"""This module defines the base LR scheduler."""

__all__ = ["BaseLRScheduler", "LRSchedulerType", "setup_lr_scheduler"]

import logging
from typing import Optional, Union

from objectory import AbstractFactory
from torch.optim import Optimizer
from torch.optim.lr_scheduler import ReduceLROnPlateau
from torch.optim.lr_scheduler import _LRScheduler
from torch.optim.lr_scheduler import _LRScheduler as LRScheduler

from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)


class BaseLRScheduler(_LRScheduler, metaclass=AbstractFactory):
    r"""Defines the base learning rate scheduler."""


# Define this type because there is not a unique class to describe a LR scheduler.
LRSchedulerType = Union[BaseLRScheduler, ReduceLROnPlateau, _LRScheduler]


def setup_lr_scheduler(
    optimizer: Optional[Optimizer],
    lr_scheduler: Union[LRScheduler, dict, None],
) -> Optional[LRScheduler]:
    r"""Sets up a learning rate scheduler.

    Args:
        optimizer (``torch.optim.Optimizer`` or ``None``): Specifies
            the optimizer.
        lr_scheduler (``torch.optim.lr_scheduler._LRScheduler`` or dict
            or ``None``): Specifies the learning rate scheduler or its
            configuration.

    Returns:
        ``torch.optim.lr_scheduler._LRScheduler`` or ``None``: An
            instantiated learning rate scheduler or ``None`` if it is
            not possible to instantiate a learning rate scheduler.
    """
    if lr_scheduler is None:
        logger.info("No LR scheduler")
        return None
    if optimizer is None:
        logger.info("The LR scheduler is not initialized because there is no optimizer")
        return None
    if isinstance(lr_scheduler, dict):
        logger.info(
            "Initializing a LR scheduler from its configuration... "
            f"{str_target_object(lr_scheduler)}"
        )
        lr_scheduler = BaseLRScheduler.factory(optimizer=optimizer, **lr_scheduler)
    if hasattr(lr_scheduler, "optimizer") and lr_scheduler.optimizer != optimizer:
        logger.warning(
            "The optimizer associated to the LR scheduler is different from the optimizer. "
            "Please verify that the LR scheduler was correctly initialized with the right "
            "optimizer."
        )
    return lr_scheduler



__all__ = [
    "BaseEngineSaver",
    "BaseHandler",
    "BestEngineStateSaver",
    "BestHistorySaver",
    "ConsolidateOptimizerState",
    "EarlyStopping",
    "EngineStateLoader",
    "EngineStateLoaderWithExcludeKeys",
    "EngineStateLoaderWithIncludeKeys",
    "EpochCudaMemoryMonitor",
    "EpochEngineStateSaver",
    "EpochLRMonitor",
    "EpochLRScheduler",
    "EpochLRSchedulerUpdater",
    "EpochOptimizerMonitor",
    "EpochSysInfoMonitor",
    "IterationCudaMemoryMonitor",
    "IterationLRMonitor",
    "IterationLRScheduler",
    "IterationLRSchedulerUpdater",
    "IterationOptimizerMonitor",
    "LRSchedulerUpdater",
    "LastHistorySaver",
    "MetricEpochLRSchedulerUpdater",
    "MetricLRSchedulerUpdater",
    "ModelArchitectureAnalyzer",
    "ModelNetworkArchitectureAnalyzer",
    "ModelParameterAnalyzer",
    "ModelStateDictLoader",
    "ParameterInitializer",
    "PartialModelStateDictLoader",
    "TagEngineStateSaver",
    "VanillaLRScheduler",
    "add_unique_event_handler",
    "setup_and_attach_handlers",
    "setup_handler",
    "to_events",
]

from gravitorch.handlers.base import BaseHandler
from gravitorch.handlers.cudamem import (
    EpochCudaMemoryMonitor,
    IterationCudaMemoryMonitor,
)
from gravitorch.handlers.early_stopping import EarlyStopping
from gravitorch.handlers.engine_loader import (
    EngineStateLoader,
    EngineStateLoaderWithExcludeKeys,
    EngineStateLoaderWithIncludeKeys,
)
from gravitorch.handlers.engine_saver import (
    BaseEngineSaver,
    BestEngineStateSaver,
    BestHistorySaver,
    EpochEngineStateSaver,
    LastHistorySaver,
    TagEngineStateSaver,
)
from gravitorch.handlers.lr_monitor import EpochLRMonitor, IterationLRMonitor
from gravitorch.handlers.lr_scheduler import (
    EpochLRScheduler,
    IterationLRScheduler,
    VanillaLRScheduler,
)
from gravitorch.handlers.lr_scheduler_updater import (
    EpochLRSchedulerUpdater,
    IterationLRSchedulerUpdater,
    LRSchedulerUpdater,
    MetricEpochLRSchedulerUpdater,
    MetricLRSchedulerUpdater,
)
from gravitorch.handlers.model_architecture_analyzer import (
    ModelArchitectureAnalyzer,
    ModelNetworkArchitectureAnalyzer,
)
from gravitorch.handlers.model_parameter_analyzer import ModelParameterAnalyzer
from gravitorch.handlers.model_state_dict_loader import (
    ModelStateDictLoader,
    PartialModelStateDictLoader,
)
from gravitorch.handlers.optimizer_monitor import (
    EpochOptimizerMonitor,
    IterationOptimizerMonitor,
)
from gravitorch.handlers.optimizer_state import ConsolidateOptimizerState
from gravitorch.handlers.parameter_initializer import ParameterInitializer
from gravitorch.handlers.sysinfo import EpochSysInfoMonitor
from gravitorch.handlers.utils import (
    add_unique_event_handler,
    setup_and_attach_handlers,
    setup_handler,
    to_events,
)



__all__ = ["ExponentialNLLLoss"]

from torch import Tensor
from torch.nn import Module

from gravitorch.nn.functional.experimental import exponential_nll_loss
from gravitorch.nn.functional.loss_helpers import check_basic_loss_reduction


class ExponentialNLLLoss(Module):
    r"""Implements a criterion to compute the negative log-likelihood (NLL) with
    Exponential distribution of target.

    Args:
        log_input (bool, optional): If ``True``, the expected input
            is ``log(lambda)``, otherwise it is ``lambda``.
            Default: ``True``
        eps (float, optional): Small value to avoid evaluation of
            :math:`\log(0)` when :attr:`log_input = False`.
            Default: ``1e-8``
        max_log_value (float, optional): Specifies the maximum value
            used to clip ``log_rate`` before to compute the
            exponential when :attr:`log_input = True`.
            Default: ``20.0``
        reduction (string, optional): Specifies the reduction to apply
            to the output: ``'none'`` | ``'mean'`` | ``'sum'``.
            ``'none'``: no reduction will be applied, ``'mean'``: the
            sum of the output will be divided by the number of
            elements in the output, ``'sum'``: the output will be
            summed. Default: ``'mean'``
    """

    def __init__(
        self,
        log_input: bool = True,
        eps: float = 1e-8,
        max_log_value: float = 20.0,
        reduction: str = "mean",
    ):
        super().__init__()
        self._log_input = bool(log_input)

        if eps < 0:
            raise ValueError(f"eps has to be greater or equal to 0 but received {eps}")
        self._eps = float(eps)
        self._max_log_value = float(max_log_value)

        check_basic_loss_reduction(reduction)
        self.reduction = str(reduction)

    def extra_repr(self) -> str:
        return (
            f"log_input={self._log_input}, eps={self._eps}, "
            f"max_log_value={self._max_log_value}, reduction={self.reduction}"
        )

    def forward(self, log_rate: Tensor, target: Tensor) -> Tensor:
        r"""Computes the negative log-likelihood (NLL) with Exponential
        distribution of target.

        Args:
            log_rate (``torch.Tensor`` of type float): Specifies the
                predicted rates (``lambda``).
            target (``torch.Tensor`` of type float and same shape as
                ``log_rate``): Specifies the target values.

        Returns:
            ``torch.Tensor`` of type float: The negative
                log-likelihood with Exponential distribution of
                target. The shape of the tensor depends on the
                reduction strategy.
        """
        return exponential_nll_loss(
            log_rate=log_rate,
            target=target,
            log_input=self._log_input,
            eps=self._eps,
            max_log_value=self._max_log_value,
            reduction=self.reduction,
        )



__all__ = ["EngineEvents"]


class EngineEvents:
    r"""Engine specific event names.

    Every engine should fire these events.
    """
    # Generic. These events are used in both ``train`` and ``eval``
    STARTED: str = "started"
    EPOCH_STARTED: str = "epoch_started"
    EPOCH_COMPLETED: str = "epoch_completed"
    COMPLETED: str = "completed"

    # Train
    TRAIN_STARTED: str = "train_started"
    TRAIN_EPOCH_STARTED: str = "train_epoch_started"
    TRAIN_ITERATION_STARTED: str = "train_iteration_started"
    TRAIN_FORWARD_COMPLETED: str = "train_forward_completed"
    TRAIN_BACKWARD_COMPLETED: str = "train_backward_completed"
    TRAIN_ITERATION_COMPLETED: str = "train_iteration_completed"
    TRAIN_EPOCH_COMPLETED: str = "train_epoch_completed"
    TRAIN_COMPLETED: str = "train_completed"

    # Eval
    EVAL_STARTED: str = "eval_started"
    EVAL_EPOCH_STARTED: str = "eval_epoch_started"
    EVAL_ITERATION_STARTED: str = "eval_iteration_started"
    EVAL_ITERATION_COMPLETED: str = "eval_iteration_completed"
    EVAL_EPOCH_COMPLETED: str = "eval_epoch_completed"
    EVAL_COMPLETED: str = "eval_completed"



__all__ = ["log_box_dataset_class"]

import logging

from torch.utils.data import Dataset

logger = logging.getLogger(__name__)


def log_box_dataset_class(dataset: Dataset) -> None:
    r"""Logs the name of the dataset in a "box".

    Args:
        dataset (``torch.utils.data.Dataset``): The dataset to log.

    Example usage:

    .. code-block:: python

        >>> import torchvision
        >>> from gravitorch.data.datasets import log_box_dataset_class
        >>> dataset = torchvision.datasets.MNIST(...)
        >>> log_box_dataset_class(dataset)
         -------
        | MNIST |
         -------
    """
    dataset_name = dataset.__class__.__qualname__
    logger.info(" " + "-" * (len(dataset_name) + 2))
    logger.info("| " + dataset_name + " |")
    logger.info(" " + "-" * (len(dataset_name) + 2))



r"""This module implements some utility functions for the training loops."""

__all__ = ["setup_training_loop"]

import logging
from typing import Union

from gravitorch.utils.format import str_target_object
from gravitorch.utils.training_loops.base import BaseTrainingLoop
from gravitorch.utils.training_loops.vanilla import VanillaTrainingLoop

logger = logging.getLogger(__name__)


def setup_training_loop(training_loop: Union[BaseTrainingLoop, dict, None]) -> BaseTrainingLoop:
    r"""Sets up the training loop.

    The training loop is instantiated from its configuration by using
    the ``BaseTrainingLoop`` factory function.

    Args:
        training_loop (``BaseTrainingLoop`` or dict or None):
            Specifies the training loop or its configuration.
            If ``None``, the ``VanillaTrainingLoop`` is instantiated.

    Returns:
        ``BaseTrainingLoop``: The training loop.
    """
    if training_loop is None:
        training_loop = VanillaTrainingLoop()
    if isinstance(training_loop, dict):
        logger.info(
            "Initializing a training loop from its configuration... "
            f"{str_target_object(training_loop)}"
        )
        training_loop = BaseTrainingLoop.factory(**training_loop)
    return training_loop



__all__ = [
    "BaseResourceRunner",
    "BaseRunner",
    "EvaluationRunner",
    "NoRepeatRunner",
    "TrainingRunner",
    "configure_pytorch",
    "setup_runner",
]

from gravitorch.runners.base import BaseRunner
from gravitorch.runners.evaluation import EvaluationRunner
from gravitorch.runners.no_repeat import NoRepeatRunner
from gravitorch.runners.resource import BaseResourceRunner
from gravitorch.runners.training import TrainingRunner
from gravitorch.runners.utils import configure_pytorch, setup_runner



r"""This module defines some collators that use packed sequence (https:/

/pytorch.org/docs/stable/generated/torch.nn.utils.rnn.PackedSequence.htm
l) to deal with variable size data.
"""

__all__ = ["PackedSequenceCollator", "DictPackedSequenceCollator"]

from collections.abc import Hashable, Sequence

from torch.nn.utils.rnn import pack_sequence
from torch.utils.data.dataloader import default_collate

from gravitorch import constants as ct
from gravitorch.data.dataloaders.collators.base import BaseCollator


class PackedSequenceCollator(BaseCollator[tuple[dict, dict], dict]):
    """Implements a collator to create batch of packed sequences because the
    default PyTorch collate function does not work with sequences of different
    lenghts.

    With this collator, each example in the batch can have a different
    length. The examples have to be formatted in a specified format.
    The example should be a tuple with two items. The first item is a
    dict containing the fixed size inputs i.e. the input that have the
    same size for each example of the mini-batch. The first item needs
    to contain the length of the sequence because this information is
    used to sort the examples in the batch by descending order of
    sequence length. The length of the sequence is indicated by the
    key ``'length'``. The first item is combined by the default
    PyTorch collate function. The second item is a dict containing all
    the inputs that have variable size. Each variable size input
    should be of size ``L x *``, where `L` is the length of a sequence
    and `*` is any number of trailing dimensions, including zero. This
    item is combined by using packed sequences. This collator does not
    work if all the sequences are empty.

    Note that every empty sequence if removed.

    Args:
        length_key (``Hashable``, optional): Specifies the key with
            the length of each example. Default: ``'length'``
    """

    def __init__(self, length_key: Hashable = ct.LENGTH):
        self._length_key = length_key

    def __call__(self, data: list[tuple[dict, dict]]) -> dict:
        r"""Creates a batch of packed sequence examples given a list of
        examples.

        Args:
            data (list): The list of examples.

        Returns:
            dict: The generated mini-batch.
        """
        # Sort the examples by the length of their sequence.
        data.sort(key=lambda x: x[0][self._length_key], reverse=True)

        fixed_size_data, variable_size_data = zip(*data)
        # Create the batch with the fixed size inputs and remove examples with empty sequences.
        batch = default_collate(fixed_size_data)
        lengths = batch[self._length_key]
        num_seq = (lengths > 0).sum().item()

        # Remove the empty sequences.
        for key, value in batch.items():
            batch[key] = value[:num_seq]

        # Create the packed sequences.
        for key in variable_size_data[0].keys():
            batch[key] = pack_sequence([variable_size_data[i][key] for i in range(num_seq)])

        return batch

    def __str__(self) -> str:
        return f"{self.__class__.__qualname__}(length_key={self._length_key})"


class DictPackedSequenceCollator(BaseCollator[dict, dict]):
    r"""Implements a collator to create batch of packed sequences because the
    default PyTorch collate function does not work with sequences of different
    lengths.

    With this collator, each example in the batch can have a different
    length. Each example has to be a dict, and you need to specify the
    list of keys that will be aggregated by using the function
    ``pack_sequence`` of PyTorch. Each input to pack should be of size
    ``L x *``, where `L` is the length of a sequence and `*` is any
    number of trailing dimensions, including zero.

    .. note::

        This collator will not raise an error if you specify a key
        that is not in the keys of the batch.

    Args:
        keys_to_pack (``Sequence``): Specifies the sequence of keys
            to pack. The first key is used to sort the examples by
            length.
    """

    def __init__(self, keys_to_pack: Sequence[Hashable]):
        self._keys_to_pack = tuple(keys_to_pack)

    def __call__(self, data: Sequence[dict]) -> dict:
        r"""Creates a batch of packed sequence examples given a list of
        examples.

        Args:
            data (``Sequence``): The sequence of examples.

        Returns:
            dict: The generated mini-batch.
        """
        # Remove the empty sequences.
        data = [example for example in data if example[self._keys_to_pack[0]].shape[0] > 0]
        num_seq = len(data)

        # Sort the examples by the length of their sequence.
        data.sort(key=lambda x: x[self._keys_to_pack[0]].shape[0], reverse=True)

        # Create the batch of examples.
        batch = {}
        for key in data[0].keys():
            examples = [data[i][key] for i in range(num_seq)]
            if key in self._keys_to_pack:
                batch[key] = pack_sequence(examples)
            else:
                batch[key] = default_collate(examples)

        return batch

    def __str__(self) -> str:
        return f"{self.__class__.__qualname__}(keys_to_pack={self._keys_to_pack})"



__all__ = [
    "LazyFlattedTensor",
    "UNKNOWN",
    "get_dtype",
    "get_shape",
    "has_name",
    "isymlog",
    "isymlog_",
    "partial_transpose_dict",
    "permute",
    "recursive_apply",
    "recursive_contiguous",
    "recursive_detach",
    "recursive_transpose",
    "safeexp",
    "safelog",
    "scalable_quantile",
    "shapes_are_equal",
    "str_full_tensor",
    "symlog",
    "symlog_",
    "to_tensor",
]

from gravitorch.utils.tensor.flatted import LazyFlattedTensor
from gravitorch.utils.tensor.math_ops import (
    isymlog,
    isymlog_,
    safeexp,
    safelog,
    scalable_quantile,
    symlog,
    symlog_,
)
from gravitorch.utils.tensor.misc import (
    has_name,
    partial_transpose_dict,
    permute,
    shapes_are_equal,
    str_full_tensor,
    to_tensor,
)
from gravitorch.utils.tensor.recursive_ops import (
    UNKNOWN,
    get_dtype,
    get_shape,
    recursive_apply,
    recursive_contiguous,
    recursive_detach,
    recursive_transpose,
)
